"""Defines a loader mixin to load xarray objects"""

import xarray as xr

from ..containers import XrDataContainer, PassthroughContainer
from ._tools import add_loader

# -----------------------------------------------------------------------------

class XarrayLoaderMixin:
    """Supplies functionality to load xarray objects"""

    @add_loader(TargetCls=XrDataContainer)
    def _load_xr_dataarray(filepath: str, *, TargetCls: type,
                           load_completely: bool=False, **load_kwargs
                           ) -> XrDataContainer:
        """Loads an xr.DataArray from a netcdf file into an XrDataContainer.
        
        Args:
            filepath (str): Where the xarray-dumped netcdf file is located
            TargetCls (type): The class constructor
            load_completely (bool, optional): If true, will call .load() on the
                loaded DataArray to load it completely into memory
            **load_kwargs: Passed on to xr.load_dataarray, see there for kwargs
        
        Returns:
            XrDataContainer: The reconstructed XrDataContainer
        """
        da = xr.load_dataarray(filepath, **load_kwargs)

        if load_completely:
            da = da.load()

        # Create the XrDataContainer, carrying over attributes
        return TargetCls(data=da, attrs=da.attrs)

    @add_loader(TargetCls=PassthroughContainer)
    def _load_xr_dataset(filepath: str, *, TargetCls: type,
                         load_completely: bool=False, **load_kwargs
                         ) -> PassthroughContainer:
        """Loads an xr.Dataset from a netcdf file into a PassthroughContainer.

        .. note::
            As there is no proper equivalent of a dataset in dantro (yet), and
            unpacking the dataset into a dantro group would reduce
            functionality, the PassthroughContainer is used here. It should
            behave almost the same as an xr.Dataset.
        
        Args:
            filepath (str): Where the xarray-dumped netcdf file is located
            TargetCls (type): The class constructor
            load_completely (bool, optional): If true, will call .load() on the
                loaded xr.Dataset to load it completely into memory.
            **load_kwargs: Passed on to xr.load_dataarray, see there for kwargs
        
        Returns:
            PassthroughContainer: The reconstructed XrDataset, stored in a
                passthrough container.
        """
        ds = xr.load_dataset(filepath, **load_kwargs)

        if load_completely:
            ds = ds.load()

        # Create the PassthroughContainer, carrying over dataset attributes
        return TargetCls(data=ds, attrs=ds.attrs)



"""This is an implementation of a DAG for transformations on dantro objects"""

import os
import sys
import glob
import copy
import time
import logging
import pickle as pkl
from itertools import chain

from typing import TypeVar, Dict, Tuple, Sequence, Any, Union, List, Set

import numpy as np
import xarray as xr

from paramspace.tools import recursive_replace, recursive_collect

from .abc import AbstractDataContainer, PATH_JOIN_CHAR
from .base import BaseDataGroup
from .utils import KeyOrderedDict, apply_operation, register_operation
from .tools import recursive_update
from .data_loaders import LOADER_BY_FILE_EXT
from .containers import ObjectContainer, NumpyDataContainer, XrDataContainer
from ._dag_utils import (DAGObjects, DAGReference, DAGTag, DAGNode,
                         parse_dag_syntax as _parse_dag_syntax,
                         parse_dag_minimal_syntax as _parse_dag_minimal_syntax)
from ._hash import _hash, SHORT_HASH_LENGTH, FULL_HASH_LENGTH


# Local constants .............................................................
log = logging.getLogger(__name__)

# The path within the DAG's associated DataManager to which caches are loaded
DAG_CACHE_DM_PATH = 'cache/dag'

# Functions that can store the DAG computation result objects, distinguishing
# by their type.
DAG_CACHE_RESULT_SAVE_FUNCS = {
    # Saving functions of specific dantro objects
    (NumpyDataContainer,):
        lambda obj, p, **kws: obj.save(p+".npy", **kws),
    (XrDataContainer,):
        lambda obj, p, **kws: obj.save(p+".xrdc", **kws),

    # Saving functions of external packages
    (np.ndarray,):
        lambda obj, p, **kws: np.save(p+".npy", obj, **kws),
    (xr.DataArray,):
        lambda obj, p, **kws: obj.to_netcdf(p+".nc_da", **kws),
    (xr.Dataset,):
        lambda obj, p, **kws: obj.to_netcdf(p+".nc_ds", **kws),
}


# -----------------------------------------------------------------------------

class Transformation:
    """A transformation is the collection of an N-ary operation and its inputs.

    Transformation objects store the name of the operation that is to be
    carried out and the arguments that are to be fed to that operation. After
    a Transformation is defined, the only interaction with them is via the
    ``compute`` method.

    For computation, the arguments are recursively inspected for whether there
    are any DAGReference-derived objects; these need to be resolved first,
    meaning they are looked up in the DAG's object database and -- if they are
    another Transformation object -- their result is computed. This can lead
    to a traversal along the DAG.

    .. warning::

        Objects of this class should under *no* circumstances be changed after
        they were created! For performance reasons, the
        :py:attr:`~dantro.dag.Transformation.hashstr` property is cached; thus,
        changing attributes that are included into the hash computation will
        not lead to a new hash, thus silently creating wrong behaviour.

        All relevant attributes (operation, args, kwargs, salt) are thus set
        read-only. This should be respected!
    """

    def __init__(self, *, operation: str,
                 args: Sequence[Union[DAGReference, Any]],
                 kwargs: Dict[str, Union[DAGReference, Any]],
                 dag: 'TransformationDAG'=None,
                 salt: int=None,
                 file_cache: dict=None):
        """Initialize a Transformation object.
        
        Args:
            operation (str): The operation that is to be carried out.
            args (Sequence[Union[DAGReference, Any]]): Positional arguments
                for the operation.
            kwargs (Dict[str, Union[DAGReference, Any]]): Keyword arguments
                for the operation. These are internally stored as a
                :py:class:`~dantro.utils.ordereddict.KeyOrderedDict`.
            dag (TransformationDAG, optional): An associated DAG that is needed
                for object lookup. Without an associated DAG, args or kwargs
                may NOT contain any object references.
            salt (int, optional): A hashing salt that can be used to let this
                specific Transformation object have a different hash than other
                objects, thus leading to cache misses.
            file_cache (dict, optional): File cache options. Expected keys are
                ``write`` (boolean or dict) and ``read`` (boolean or dict).
                Note that the options given here are NOT reflected in the hash
                of the object!

                The following arguments are possible under the ``read`` key:

                    enabled (bool, optional):
                        Whether it should be attempted to read from the file
                        cache.
                    load_options (dict, optional):
                        Passed on to the method that loads the cache,
                        :py:meth:`~dantro.data_mngr.DataManager.load`.

                Under the ``write`` key, the following arguments are possible.
                They are evaluated in the order that they are listed here.
                See :py:meth:`~dantro.dag.Transformation._cache_result` for
                more information.

                    enabled (bool, optional):
                        Whether writing is enabled at all
                    always (bool, optional):
                        If given, will always write.
                    allow_overwrite (bool, optional):
                        If False, will not write a cache file if one already
                        exists. If True, a cache file *might* be written,
                        although one already exists. This is still conditional
                        on the evaluation of the other arguments.
                    min_size (int, optional):
                        The *minimum* size of the result object that allows
                        writing the cache.
                    max_size (int, optional):
                        The *maximum* size of the result object that allows
                        writing the cache.
                    min_compute_time (float, optional):
                        The minimal individual computation time of this node
                        that is needed in order for the file cache to be
                        written.
                        *Note* that this value can be lower if the node result
                        is not computed but looked up from the cache.
                    min_cumulative_compute_time (float, optional):
                        The minimal cumulative computation time of this node
                        and all its dependencies that is needed in order for
                        the file cache to be written.
                        *Note* that this value can be lower if the node result
                        is not computed but looked up from the cache.
                    storage_options (dict, optional):
                        Passed on to the cache storage method,
                        :py:meth:`dantro.dag.TransformationDAG._write_to_cache_file`.
                        The following arguments are available:

                        ignore_groups (bool, optional):
                            Whether to store groups. Disabled by default.
                        attempt_pickling (bool, optional):
                            Whether it should be attempted to store results
                            that could not be stored via a dedicated storage
                            function by pickling them. Enabled by default.
                        raise_on_error (bool, optional):
                            Whether to raise on error to store a result.
                            Disabled by default; it is useful to enable this
                            when debugging.
                        pkl_kwargs (dict, optional):
                            Arguments passed on to the pickle.dump function.
                        further keyword arguments:
                            Passed on to the chosen storage method.
        """
        # Storage attributes
        self._operation = operation
        self._args = args
        self._kwargs = KeyOrderedDict(**kwargs)
        self._dag = dag
        self._salt = salt
        self._hashstr = None
        self._profile = dict(compute=0., cumulative_compute=0.,
                             hashstr=0., cache_lookup=0., cache_writing=0.)

        # Parse file cache options, making sure it's a dict with default values
        self._fc_opts = file_cache if file_cache is not None else {}
        
        if isinstance(self._fc_opts.get('write', {}), bool):
            self._fc_opts['write'] = dict(enabled=self._fc_opts['write'])
        elif 'write' not in self._fc_opts:
            self._fc_opts['write'] = dict(enabled=False)
        
        if isinstance(self._fc_opts.get('read', {}), bool):
            self._fc_opts['read'] = dict(enabled=self._fc_opts['read'])
        elif 'read' not in self._fc_opts:
            self._fc_opts['read'] = dict(enabled=False)

        # Cache dict, containing the result and whether the cache is in memory
        self._cache = dict(result=None, filled=False)

    # .........................................................................
    # String representation and hashing

    def __str__(self) -> str:
        """A human-readable string characterizing this Transformation"""
        return ("<{t:}, operation: {op:}, {Na:d} args, {Nkw:d} kwargs>\n"
                "  args:   {args:}\n"
                "  kwargs: {kwargs:}\n"
                "".format(t=type(self).__name__, op=self._operation,
                          Na=len(self._args), Nkw=len(self._kwargs),
                          args=self._args, kwargs=self._kwargs))

    def __repr__(self) -> str:
        """A deterministic string representation of this transformation.

        .. note::

            This is also used for hash creation, thus it does not include the
            attributes that are set via the initialization arguments ``dag``
            and ``file_cache``.

        .. warning::

            Changing this method will lead to cache invalidations!
        """
        return ("<{mod:}.{t:}, operation={op:}, args={args:}, "
                "kwargs={kwargs:}, salt={salt:}>"
                "".format(mod=type(self).__module__, t=type(self).__name__,
                          op=repr(self._operation),
                          args=repr(self._args),
                          kwargs=repr(dict(self._kwargs)),# TODO Check sorting!
                          salt=repr(self._salt)))

    @property
    def hashstr(self) -> str:
        """Computes the hash of this Transformation by creating a deterministic
        representation of this Transformation using ``__repr__`` and then
        applying a checksum hash function to it.
        
        Note that this does NOT rely on the built-in hash function but on the
        custom dantro ``_hash`` function which produces a platform-independent
        and deterministic hash. As this is a *string*-based (rather than an
        integer-based) hash, it is not implemented as the ``__hash__`` magic
        method but as this separate property.
        
        Returns:
            str: The hash string for this transformation
        """
        if self._hashstr is None:
            t0 = time.time()
            self._hashstr = _hash(repr(self))
            self._update_profile(hashstr=time.time() - t0)

        return self._hashstr

    def __hash__(self) -> int:
        """Computes the python-compatible integer hash of this object from the
        string-based hash of this Transformation.
        """
        return hash(self.hashstr)

    # .........................................................................
    # Properties

    @property
    def dag(self) -> 'TransformationDAG':
        """The associated TransformationDAG; used for object lookup"""
        return self._dag

    @property
    def dependencies(self) -> Set[DAGReference]:
        """Recursively collects the references that are found in the positional
        and keyword arguments of this Transformation.
        """
        return set(recursive_collect(chain(self._args, self._kwargs.values()),
                   select_func=(lambda o: isinstance(o, DAGReference))))

    @property
    def resolved_dependencies(self) -> Set['Transformation']:
        """Transformation objects that this Transformation depends on"""
        return set([ref.resolve_object(dag=self.dag)
                    for ref in self.dependencies])

    @property
    def profile(self) -> Dict[str, float]:
        """The profiling data for this transformation"""
        return self._profile

    # YAML representation .....................................................
    yaml_tag = u'!dag_trf'

    @classmethod
    def from_yaml(cls, constructor, node):
        return cls(**constructor.construct_mapping(node, deep=True))

    @classmethod
    def to_yaml(cls, representer, node):
        """A YAML representation of this Transformation, including all its
        arguments (which must again be YAML-representable). In essence, this
        returns a YAML mapping that has the ``!dag_trf`` YAML tag prefixed,
        such that *reading* it in will lead to the ``from_yaml`` method being
        invoked.

        .. note::

            The YAML representation does *not* include the ``file_cache``
            parameters.

        .. warning::

            The YAML representation is used in computing the hashstr that
            identifies this transformation.
            Changing the argument order here or adding further keys to the
            dict will lead to hash changes and thus to cache misses.

        """
        # Collect the attributes that are relevant for the transformation.
        d = dict(operation=node._operation,
                 args=node._args,
                 kwargs=dict(node._kwargs))
        
        # If a specific salt was given, add that to the dict as well
        if node._salt is not None:
            d['salt'] = node._salt

        # Let YAML represent this as a mapping with an additional tag
        return representer.represent_mapping(cls.yaml_tag, d)

    # .........................................................................
    # Compute interface

    def compute(self) -> Any:
        """Computes the result of this transformation by recursively resolving
        objects and carrying out operations.
        
        This method can also be called if the result is already computed; this
        will lead only to a cache-lookup, not a re-computation.
        
        Returns:
            Any: The result of the operation
        """
        def is_DAGReference(obj: Any) -> bool:
            return isinstance(obj, DAGReference)

        def resolve_and_compute(ref: DAGReference):
            """Resolve references to their objects, if necessary computing the
            results of referenced Transformation objects recursively.

            Makes use of arguments from outer scope.
            """
            if self.dag is None:
                raise ValueError("Cannot resolve Transformation arguments "
                                 "that contain DAG references, because no DAG "
                                 "was associated with this Transformation!")

            # Let the reference resolve the corresponding object from the DAG
            obj = ref.resolve_object(dag=self.dag)
            
            # Check if this refers to an object that is NOT a transformation.
            # This could be the DataManager or the DAG itself, but also other
            # objects in the DAG's object database. Performing computation on
            # those is either not possible or would lead to infinite loops.
            if not isinstance(obj, Transformation):
                return obj

            # else: It is another Transformation object. Compute it, which
            # leads to a traversal up the DAG tree.
            return obj.compute()

        # Try to look up an already computed result from memory or file cache
        success, res = self._lookup_result()
        
        if not success:
            # Did not find a result in memory or file cache -> Compute it.
            # First, compute the result of the references in the arguments.
            args =   recursive_replace(copy.deepcopy(self._args),
                                       select_func=is_DAGReference,
                                       replace_func=resolve_and_compute)
            kwargs = recursive_replace(copy.deepcopy(self._kwargs),
                                       select_func=is_DAGReference,
                                       replace_func=resolve_and_compute)
            # NOTE Important to deepcopy here, because otherwise the recursive
            #      replacement and the mutability of both args and kwargs will
            #      lead to DAGReference objects being replaced with the actual
            #      objects, which would break the .hashstr property of this
            #      object. The deepcopy is always possible, because even if
            #      the args and kwargs are nested, they contain only trivial
            #      objects.

            # Carry out the operation
            res = self._perform_operation(args=args, kwargs=kwargs)

        # Allow caching the result, even if it comes from the cache
        self._cache_result(res)

        return res

    def _perform_operation(self, *, args, kwargs) -> Any:
        """Perform the operation, updating the profiling info on the side"""
        # Initialize a dict for profiling info
        prof = dict()

        # Set up profiling
        t0 = time.time()

        # Actually perform the operation
        res = apply_operation(self._operation, *args, **kwargs)
        # TODO Add error handling with node information

        # Prase profiling info and return the result
        self._update_profile(cumulative_compute=(time.time() - t0))

        return res

    def _update_profile(self, *,
                        cumulative_compute: float=None,
                        **times) -> None:
        """Given some new profiling times, updates the profiling information.
        
        Args:
            cumulative_compute (float, optional): The cumulative computation
                time; if given, additionally computes the computation time for
                this individual node.
            **times: Valid profiling data.
        """
        # If cumulative computation time was given, calculate individual time
        if cumulative_compute is not None:
            self._profile['cumulative_compute'] = cumulative_compute

            # Aggregate the dependencies' cumulative computation times
            deps_cctime = sum([dep.profile['cumulative_compute']
                               for dep in self.resolved_dependencies
                               if isinstance(dep, Transformation)])
            # NOTE The dependencies might not have this value set because there
            #      might have been a cache lookup
            self._profile['compute'] = max(0., cumulative_compute-deps_cctime)

        # Store the remaining entries
        self._profile.update(times)

    # Cache handling . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    def _lookup_result(self) -> Tuple[bool, Any]:
        """Look up the transformation result to spare re-computation"""
        success, res = False, None

        # Retrieve cache parameters
        read_opts = self._fc_opts.get('read', {})
        load_opts = read_opts.get('load_options', {})

        # Check if the cache is already filled. If not, see if the file cache
        # can be read and is configured to be read.
        if self._cache['filled']:
            success = True
            res = self._cache['result']

        elif self.dag is not None and read_opts.get('enabled', False):
            # Setup profiling
            t0 = time.time()

            # Let the DAG check if there is a file cache, i.e. if a file with
            # this Transformation's hash exists in the DAG's cache directory.
            success, res = self.dag._retrieve_from_cache_file(self.hashstr,
                                                              **load_opts)

            # Store the result
            if success:
                self._cache['result'] = res
                self._cache['filled'] = True

            self._update_profile(cache_lookup=(time.time() - t0))

        return success, res
    
    def _cache_result(self, result: Any) -> None:
        """Stores a computed result in the cache"""
        def should_write(*, enabled: bool, always: bool=False,
                         allow_overwrite: bool=False,
                         min_size: int=None, max_size: int=None,
                         min_compute_time: float=None,
                         min_cumulative_compute_time: float=None,
                         storage_options: dict=None
                         ) -> bool:
            """A helper function to evaluate _whether_ the file cache is to be
            written or not.
            
            Args:
                enabled (bool): Whether writing is enabled at all
                always (bool, optional): If given, will always write.
                allow_overwrite (bool, optional): If False, will not write a
                    cache file if one already exists. If True, a cache file
                    _might_ be written, although one already exists. This is
                    still conditional on the evaluation of the other arguments.
                min_size (int, optional): The minimum size of the result object
                    that allows writing the cache.
                max_size (int, optional): The maximum size of the result object
                    that allows writing the cache.
                min_compute_time (float, optional): The minimal individual
                    computation time of this node that is needed in order for
                    the file cache to be written. Note that this value can be
                    lower if the node result is not computed but looked up from
                    the cache.
                min_cumulative_compute_time (float, optional): The minimal
                    cumulative computation time of this node and all its
                    dependencies that is needed in order for the file cache to
                    be written. Note that this value can be lower if the node
                    result is not computed but looked up from the cache.
                storage_options (dict, optional): (ignored here)
            
            Returns:
                bool: Whether to write the file cache or not.
            """
            if not enabled:
                # ... nothing else to check
                return False

            # With always: always write, don't look at other arguments.
            if always:
                return True
            # All checks below are formulated such that they return False.

            # If overwriting is _disabled_ and a cache file already exists, it
            # is already clear that a new one should _not_ be written
            if not allow_overwrite and self.hashstr in self.dag.cache_files:
                return False

            # Evaluate profiling information
            if min_compute_time is not None:
                if self.profile['compute'] < min_compute_time:
                    return False

            if min_cumulative_compute_time is not None:
                if (  self.profile['cumulative_compute']
                    < min_cumulative_compute_time):
                    return False

            # Evaluate object size
            if min_size is not None or max_size is not None:
                size_itvl = [min_size if min_size is not None else 0,
                             max_size if max_size is not None else np.inf]
                obj_size = sys.getsizeof(result)  # from outer scope

                if not (size_itvl[0] < obj_size < size_itvl[1]):
                    return False

            # If this point is reached, the cache file should be written.
            return True

        # Store a reference to the result and mark the cache as being in use
        self._cache['result'] = result
        self._cache['filled'] = True
        # NOTE If instead of a proper computation, the passed result object was
        #      previously looked up from the cache, this will not have an
        #      effect.
        
        # Get file cache writing parameters; don't write if not 
        write_opts = self._fc_opts['write']

        # Determine whether to write to a file
        if self.dag is not None and should_write(**write_opts):
            # Setup profiling
            t0 = time.time()
            
            # Write the result to a file inside the DAG's cache directory. This
            # is handled by the DAG itself, because the Transformation does not
            # know (and should not care) aboute the cache directory ...
            storage_opts = write_opts.get('storage_options', {})
            self.dag._write_to_cache_file(self.hashstr, result=result,
                                          **storage_opts)

            self._update_profile(cache_writing=(time.time() - t0))



# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

class TransformationDAG:
    """This class collects transformation operations that are (already by
    their own structure) connected into a directed acyclic graph. The aim of
    this class is to maintain base objects, manage references, and allow
    operations on the DAG, the most central of which is computing the result
    of a node.

    Furthermore, this class also implements caching of transformations, such
    that operations that take very long can be stored (in memory or on disk) to
    speed up future operations.

    Objects of this class are initialized with dict-like arguments which
    specify the transformation operations. There are some shorthands that allow
    a simple definition syntax, for example the ``select`` syntax, which takes
    care of selecting a basic set of data from the associated DataManager.
    """

    def __init__(self, *, dm: 'DataManager',
                 select: dict=None, transform: Sequence[dict]=None,
                 cache_dir: str='.cache', file_cache_defaults: dict=None,
                 base_transform: Sequence[Transformation]=None,
                 select_base: Union[DAGReference, str]=None,
                 select_path_prefix: str=None):
        """Initialize a DAG which is associated with a DataManager and load the
        specified transformations configuration into it.
        
        Args:
            dm (DataManager): The associated data manager
            select (dict, optional): Selection specifications, which are
                translated into regular transformations based on ``getitem``
                operations. The ``base_transform`` and ``select_base``
                arguments can be used to define from which object to select.
                By default, selection happens from the associated DataManager.
            transform (Sequence[dict], optional): Transform specifications.
            cache_dir (str, optional): The name of the cache directory to
                create if file caching is enabled. If this is a relative path,
                it is interpreted relative to the associated data manager's
                data directory. If it is absolute, the absolute path is used.
                The directory is only created if it is needed.
            file_cache_defaults (dict, optional): Default arguments for file
                caching behaviour. This is recursively updated with the
                arguments given in each individual select or transform
                specification.
            base_transform (Sequence[Transformation], optional): A sequence of
                transform specifications that are added to the DAG prior to
                those added via ``select`` and ``transform``. These can be used
                to create some other object from the data manager which should
                be used as the basis of ``select`` operations.
            select_base (Union[DAGReference, str], optional): Which tag to
                base the ``select`` operations on. If None, will use the
                (always-registered) tag for the data manager, ``dm``. This
                attribute can also be set via the ``select_base`` property.
            select_path_prefix (str, optional): If given, this path is prefixed
                to all ``path`` specifications made within the ``select``
                argument. Note that unlike setting the ``select_base`` this
                merely joins the given prefix to the given paths, thus leading
                to repeated path resolution. For that reason, using the
                ``select_base`` argument is generally preferred and the
                ``select_path_prefix`` should only be used if ``select_base``
                is already in use.
                If this path ends with a ``/``, it is directly prepended. If
                not, the ``/`` is added before adjoining it to the other path.
        """
        self._dm = dm
        self._objects = DAGObjects()
        self._tags = dict()
        self._nodes = list()
        self._fc_opts = file_cache_defaults if file_cache_defaults else {}
        self._select_base = None
        self._profile = dict(add_node=0., compute=0.)
        self._select_path_prefix = select_path_prefix
        
        # Determine cache directory path; relative path interpreted as relative
        # to the DataManager's data directory
        if os.path.isabs(cache_dir):
            self._cache_dir = cache_dir
        else:
            self._cache_dir = os.path.join(self.dm.dirs['data'], cache_dir)

        # Add the DAG itself and the DataManager as objects with default tags
        self.tags['dag'] = self.objects.add_object(self)
        self.tags['dm'] = self.objects.add_object(self.dm)
        # NOTE The data manager is NOT a node of the DAG, but more like an
        #      external data source, thus being accessible only as a tag

        # Add base transformations that do not rely on select operations
        self.add_nodes(transform=base_transform)

        # Set the selection base tag; the property setter checks availability
        self.select_base = select_base

        # Now add nodes via the main arguments; these can now make use of the
        # select interface, because a select base tag is set and base transform
        # operations were already added.
        self.add_nodes(select=select, transform=transform)

    # .........................................................................
    
    def __str__(self) -> str:
        """A human-readable string characterizing this TransformationDAG"""
        return ("<TransformationDAG, "
                "{:d} node(s), {:d} tag(s), {:d} object(s)>"
                "".format(len(self.nodes), len(self.tags), len(self.objects)))

    # .........................................................................

    @property
    def dm(self) -> 'DataManager':
        """The associated DataManager"""
        return self._dm

    @property
    def hashstr(self) -> str:
        """Returns the hash of this DAG, which depends solely on the hash of
        the associated DataManager.
        """
        return _hash("<TransformationDAG, coupled to DataManager with ref {}>"
                     "".format(self.dm.hashstr))

    @property
    def objects(self) -> DAGObjects:
        """The object database"""
        return self._objects

    @property
    def tags(self) -> Dict[str, str]:
        """A mapping from tags to objects' hashes; the hashes can be looked
        up in the object database to get to the objects.
        """
        return self._tags

    @property
    def nodes(self) -> List[str]:
        """The nodes of the DAG"""
        return self._nodes

    @property
    def cache_dir(self) -> str:
        """The path to the cache directory that is associated with the
        DataManager that is coupled to this DAG. Note that the directory might
        not exist yet!
        """
        return self._cache_dir

    @property
    def cache_files(self) -> Dict[str, Tuple[str, str]]:
        """Scans the cache directory for cache files and returns a dict that
        has as keys the hash strings and as values a tuple of full path and
        file extension.
        """
        info = dict()

        # Go over all files in the cache dir that have an extension
        for path in glob.glob(os.path.join(self.cache_dir, '*.*')):
            if not os.path.isfile(path):
                continue

            # Get filename and extension, then check if it is a hash
            fname, ext = os.path.splitext(os.path.basename(path))
            if len(fname) != FULL_HASH_LENGTH:
                continue
            # else: filename is assumed to be the hash.

            if fname in info:
                raise ValueError("Encountered a duplicate cache file for the "
                                 "transformation with hash {}! Delete all but "
                                 "one of those files from the cache directory "
                                 "{}.".format(fname, self.cache_dir))

            # All good, store info.
            info[fname] = dict(full_path=path, ext=ext)

        return info

    @property
    def select_base(self) -> DAGReference:
        """The reference to the object that is used for select operations"""
        return self._select_base

    @select_base.setter
    def select_base(self, new_base: Union[DAGReference, str]):
        """Set the reference that is to be used as the base of select
        operations. It can either be a reference object or a string, which is
        then interpreted as a tag.
        """
        # Distinguish by type. If it's not a DAGReference, assume it's a tag.
        if new_base is None:
            new_base = DAGTag('dm').convert_to_ref(dag=self)

        elif isinstance(new_base, DAGReference):
            # Make sure it is a proper DAGReference object (hash-based) and not
            # an object of a derived class.
            new_base = new_base.convert_to_ref(dag=self)

        elif new_base not in self.tags:
            raise KeyError("The tag '{}' cannot be the basis of future select "
                           "operations because it is not available! Make sure "
                           "that a node with that tag is added prior to the "
                           "attempt of setting it. Available tags: {}. "
                           "Alternatively, pass a DAGReference object."
                           "".format(new_base, ", ".join(self.tags)))

        else:
            # Tag is available. Create a DAGReference via DAGTag conversion
            log.debug("Setting select_base to tag '%s' ...", new_base)
            new_base = DAGTag(new_base).convert_to_ref(dag=self)

        # Have a DAGReference now. Store it.
        self._select_base = new_base

    @property
    def profile(self) -> Dict[str, float]:
        """Returns the profiling information for the DAG."""
        return self._profile
    
    @property
    def profile_extended(self) -> Dict[str, Union[float, Dict[str, float]]]:
        """Builds an extended profile that includes the profiles from all
        transformations and some aggregated information.

        This is calculated anew upon each invocation; the result is not cached.
        """
        prof = copy.deepcopy(self.profile)
        
        # Add tag-specific information
        prof['tags'] = dict()
        for tag, obj_hash in self.tags.items():
            obj = self.objects[obj_hash]
            if not isinstance(obj, Transformation):
                continue

            tprof = copy.deepcopy(obj.profile)
            prof['tags'][tag] = tprof

        # Aggregate the profiled times from all transformations (by item)
        to_aggregate = ('compute', 'hashstr',
                        'cache_lookup', 'cache_writing')
        stat_funcs = dict(mean=lambda d: np.mean(d),
                          std=lambda d: np.std(d),
                          min=lambda d: np.min(d),
                          max=lambda d: np.max(d),
                          q25=lambda d: np.quantile(d, .25),
                          q50=lambda d: np.quantile(d, .50),
                          q75=lambda d: np.quantile(d, .75))
        tprofs = {item: list() for item in to_aggregate}

        for obj_hash, obj in self.objects.items():
            if not isinstance(obj, Transformation):
                continue

            tprof = copy.deepcopy(obj.profile)
            for item in to_aggregate:
                tprofs[item].append(tprof[item])

        # Compute some statistics for the aggregated elements
        prof['aggregated'] = dict()
        for item in to_aggregate:
            prof['aggregated'][item] = {k: (f(tprofs[item])
                                            if tprofs[item] else np.nan)
                                        for k, f in stat_funcs.items()}
        return prof

    # .........................................................................

    def add_node(self, *, operation: str, args: list=None, kwargs: dict=None,
                 tag: str=None, file_cache: dict=None,
                 **trf_kwargs) -> DAGReference:
        """Add a new node by creating a new Transformation object and adding it
        to the node list.
        
        Args:
            operation (str): The name of the operation
            args (list, optional): Positional arguments to the operation
            kwargs (dict, optional): Keyword arguments to the operation
            tag (str, optional): The tag the transformation should be made
                available as.
            file_cache (dict, optional): File cache options for
                this node. If defaults were given during initialization, those
                defaults will be updated with the given dict.
            **trf_kwargs: Passed on to Transformation.__init__
        
        Raises:
            ValueError: If the tag already exists
        
        Returns:
            DAGReference: The reference to the created node
        """
        t0 = time.time()

        # Some helper methods for the recursive replacement
        def not_proper_ref(obj: Any) -> bool:
            return (    isinstance(obj, DAGReference)
                    and type(obj) is not DAGReference)

        def convert_to_ref(obj: DAGReference) -> DAGReference:
            return obj.convert_to_ref(dag=self)

        # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # Handle default values of arguments
        args = copy.deepcopy(args) if args else []
        kwargs = copy.deepcopy(kwargs) if kwargs else {}
        # NOTE Deep copy is important here, because the mutability of nested
        #      args or kwargs may lead to side effects. The deep copy should
        #      always be possible, because args and kwargs should only contain
        #      trivial objects.

        # Recursively replace any derived references to proper DAGReferences,
        # which work hash-based. This is to not have multiple options of how
        # another TransformationDAG object is referenced.
        args =   recursive_replace(args, select_func=not_proper_ref,
                                   replace_func=convert_to_ref)
        kwargs = recursive_replace(kwargs, select_func=not_proper_ref,
                                   replace_func=convert_to_ref)
        
        # Parse file cache parameters
        fc_opts = copy.deepcopy(self._fc_opts)  # Always a dict
        
        if file_cache is not None:
            fc_opts = recursive_update(fc_opts, file_cache)

        # From these arguments, create the Transformation object and add it to
        # the objects database.
        trf = Transformation(operation=operation, args=args, kwargs=kwargs,
                             dag=self, file_cache=fc_opts, **trf_kwargs)
        trf_hash = self.objects.add_object(trf)
        # NOTE From this point on, the object itself has no relevance.
        #      Transformation objects should only be handled via their hash in
        #      order to reduce duplicate calculations and make efficient
        #      caching possible.

        # Store the hash in the node list
        self.nodes.append(trf_hash)

        # If a tag was specified, create a tag
        if tag:
            if tag in self.tags.keys():
                raise ValueError("Tag '{}' already exists! Choose a different "
                                 "one. Already in use: {}"
                                 "".format(tag, ", ".join(self.tags.keys())))
            self.tags[tag] = trf_hash

        # Update the profile
        self._update_profile(add_node=time.time() - t0)

        # Return a reference to the newly created node
        return DAGReference(trf_hash)
    
    def add_nodes(self, *, select: dict=None, transform: Sequence[dict]=None):
        """Adds multiple nodes by parsing the specification given via the
        ``select`` and ``transform`` arguments.
        
        Args:
            select (dict, optional): Selection specifications, which are
                translated into regular transformations based on ``getitem``
                operations. The ``base_transform`` and ``select_base``
                arguments can be used to define from which object to select.
                By default, selection happens from the associated DataManager.
            transform (Sequence[dict], optional): Transform specifications.
        """
        # Parse the arguments and add multiple nodes from those specs
        specs = self._parse_trfs(select=select, transform=transform)
        if not specs:
            return
        
        for spec in specs:
            self.add_node(**spec)
    
    def compute(self, *, compute_only: Sequence[str]=None) -> Dict[str, Any]:
        """Computes all specified tags and returns a result dict.
        
        Args:
            compute_only (Sequence[str], optional): The tags to compute. If not
                given, will compute all associated tags.
            cache_options (dict, optional): Cache options. These will update
                the default cache options given at initialization of the DAG.
        
        Returns:
            Dict[str, Any]: A mapping from tags to fully computed results.
        """
        def postprocess_result(res, *, tag: str):
            """Performs some postprocessing operations on the results of
            individual tag computations.
            """
            # If the object is a detached dantro tree object, use the short
            # transformation hash for its name
            if isinstance(res, (AbstractDataContainer)) and res.parent is None:
                res.name = trf.hashstr[:SHORT_HASH_LENGTH]

            # Unwrap ObjectContainer; these are only meant for usage within the
            # data tree and it makes little sense to keep them in that form.
            if isinstance(res, ObjectContainer):
                res = res.data

            return res
        
        # Initiate start time for profiling
        t0 = time.time()

        # Determine which tags to compute
        compute_only = compute_only if compute_only is not None else 'all'
        if compute_only == 'all':
            compute_only = [t for t in self.tags.keys()
                            if t not in ['dm', 'dag']]

        log.info("Computation invoked on DAG with %d nodes.", len(self.nodes))

        # The results dict
        results = dict()

        if not compute_only:
            log.remark("No tags were selected to be computed. "
                       "Available tags:\n  %s", ", ".join(self.tags))
            return results

        # Compute and collect the results
        for tag in compute_only:
            log.remark("Computing tag '%s' ...", tag)
            # Resolve the transformation, then compute the result
            trf = self.objects[self.tags[tag]]
            res = trf.compute()

            # Postprocess it and store the result under its tag
            results[tag] = postprocess_result(res, tag=tag)

        # Update profiling information
        t1 = time.time()
        self._update_profile(compute=t1-t0)

        # Provide some information to the user
        log.note("Computed %d tag%s in %.2gs: %s",
                 len(compute_only), "s" if len(compute_only) != 1 else "",
                 t1-t0, ", ".join(results.keys()))

        prof_extd = self.profile_extended
        fstr = ("{name:>25s}   {p[mean]:<7s}  ±  {p[std]:<7s}   "
                "({p[min]:<7s} | {p[max]:<7s})")
        log.remark("Profiling results per node:  "
                   "mean ± std (min|max) [s]\n%s",
                   "\n".join([fstr.format(name=k, p={_k: "{:.2g}".format(_v)
                                                     for _k, _v in v.items()})
                              for k, v in prof_extd['aggregated'].items()
                              if k not in ('hashstr',)]))
        # TODO In the future, make this prettier; with proper time formatting!

        return results

    # .........................................................................
    # Helpers: Parsing transformation specifications

    def _parse_trfs(self, *, select: dict,
                    transform: Sequence[dict]) -> Sequence[dict]:
        """Parse the given arguments to bring them into a uniform format: a
        sequence of parameters for transformation operations.
        
        Args:
            select (dict): The shorthand to select certain objects from the
                DataManager. These may also include transformations.
            transform (Sequence[dict]): Actual transformation operations,
                carried out afterwards.
        
        Returns:
            Sequence[dict]:
                A sequence of transformation parameters that was brought into
                a uniform structure.
        
        Raises:
            TypeError: On invalid type within entry of ``select``
            ValueError: When ``file_cache`` is given for selection from base
        """
        # The to-be-populated list of transformations
        trfs = list()

        # Prepare arguments: make sure they are dicts and deep copies.
        select = copy.deepcopy(select) if select else {}
        transform = copy.deepcopy(transform) if transform else []

        # First, parse the ``select`` argument. This contains a basic operation
        # to select data from the selection base (e.g. the DataManager) and
        # also allows to perform some operations on it.
        for tag, params in sorted(select.items()):
            if isinstance(params, str):
                path = params
                with_previous_result = False
                more_trfs = None
                salt = None
                omit_tag = False

            elif isinstance(params, dict):
                path = params['path']
                with_previous_result = params.get('with_previous_result',False)
                more_trfs = params.get('transform')
                salt = params.get('salt')
                omit_tag = params.get('omit_tag', False)

                if 'file_cache' in params:
                    raise ValueError("For selection from the selection base, "
                                     "the file cache is always disabled! "
                                     "The `file_cache` argument is thus not "
                                     "allowed; remove it from the selection "
                                     "for tag '{}'.".format(tag))

            else:
                raise TypeError("Invalid type for '{}' entry within `select` "
                                "argument! Got {} but expected string or dict."
                                "".format(tag, type(params)))

            # If given, process the path by prepending the prefix
            if self._select_path_prefix:
                if self._select_path_prefix[-1] == PATH_JOIN_CHAR:
                    path = self._select_path_prefix + path
                else:
                    path = PATH_JOIN_CHAR.join([self._select_path_prefix,path])

            # Construct parameters to select from the selection base.
            # Only assign a tag if there are no further transformations;
            # otherwise, the last additional transformation should set the tag.
            sel_trf = dict(operation='getitem',
                           tag=None if (more_trfs or omit_tag) else tag,
                           args=[self.select_base, path],
                           kwargs=dict(),
                           file_cache=dict(read=False, write=False))
            
            # Carry additional parameters only if given
            if salt is not None:
                sel_trf['salt'] = salt

            # Now finished with the formulation of the select operation.
            trfs.append(sel_trf)
            
            if not more_trfs:
                # Done with this select operation.
                continue

            # else: there are additional transformations to be parsed and added
            for i, trf_params in enumerate(more_trfs):
                trf_params = _parse_dag_minimal_syntax(trf_params)

                # Might have to use the previous result ...
                if 'with_previous_result' not in trf_params:
                    trf_params['with_previous_result'] = with_previous_result
                
                # Can now parse the regular syntax
                trf_params = _parse_dag_syntax(**trf_params)

                # If the tag is not to be omitted, the last transformation for
                # the selected tag needs to set the tag.
                if i+1 == len(more_trfs) and not omit_tag:
                    if trf_params.get('tag'):
                        raise ValueError("The tag of the last transform "
                                         "operation within a select routine "
                                         "cannot be set manually. Check the "
                                         "parameters for selection of tag "
                                         "'{}'.".format(tag))
                        # TODO Could actually allow multiple tags here ...
                    
                    # Add the tag to the parameters
                    trf_params['tag'] = tag

                # Can append it now
                trfs.append(trf_params)

        # Now, parse the normal `transform` argument. The operations defined
        # here are added after the instructions from the `select` section.
        for trf_params in transform:
            trf_params = _parse_dag_minimal_syntax(trf_params)
            trfs.append(_parse_dag_syntax(**trf_params))

        # Done parsing, yay.
        return trfs

    # .........................................................................
    # Helpers: Profiling

    def _update_profile(self, **times):
        """Updates profiling information by adding the given time to the
        matching key.
        """
        for key, t in times.items():
            self._profile[key] = self._profile.get(key, 0.) + t

    # .........................................................................
    # Cache writing and reading
    # NOTE This is done here rather than in Transformation because this is the
    #      more central entity and it is a bit easier ...

    def _retrieve_from_cache_file(self, trf_hash: str,
                                  **load_kwargs) -> Tuple[bool, Any]:
        """Retrieves a transformation's result from a cache file."""
        success, res = False, None
        cache_files = self.cache_files

        if trf_hash not in cache_files.keys():
            # Bad luck, no cache file
            return success, res

        # Parse load options
        if 'exists_action' not in load_kwargs:
            load_kwargs['exists_action'] = 'skip_nowarn'
            
        # else: There was a file. Let the DataManager load it.
        file_ext = cache_files[trf_hash]['ext']
        self.dm.load('dag_cache',
                     loader=LOADER_BY_FILE_EXT[file_ext[1:]],
                     base_path=self.cache_dir,
                     glob_str=trf_hash + file_ext,
                     target_path=DAG_CACHE_DM_PATH + "/{basename:}",
                     required=True,
                     **load_kwargs)
        # NOTE If a file was already loaded from the cache, it will not be
        #      loaded again. Thus, the DataManager acts as a persistent
        #      storage for loaded cache files. Consequently, these are shared
        #      among all TransformationDAG objects.
        
        # Retrieve from the DataManager
        res = self.dm[DAG_CACHE_DM_PATH][trf_hash]

        # Done.
        success = True
        return success, res

    def _write_to_cache_file(self, trf_hash: str, *, result: Any,
                             ignore_groups: bool=True,
                             attempt_pickling: bool=True,
                             raise_on_error: bool=False,
                             pkl_kwargs: dict=None,
                             **save_kwargs) -> bool:
        """Writes the given result object to a hash file, overwriting existing
        ones.
        
        Args:
            trf_hash (str): The hash; will be used for the file name
            result (Any): The result object to write as a cache file
            ignore_groups (bool, optional): Whether to store groups. Disabled
                by default.
            attempt_pickling (bool, optional): Whether it should be attempted
                to store results that could not be stored via a dedicated
                storage function by pickling them. Enabled by default.
            raise_on_error (bool, optional): Whether to raise on error to
                store a result. Disabled by default; it is useful to enable
                this when debugging.
            pkl_kwargs (dict, optional): Arguments passed on to the
                pickle.dump function.
            **save_kwargs: Passed on to the chosen storage method.
        
        Returns:
            bool: Whether a cache file was saved
        
        Raises:
            NotImplementedError: When attempting to store instances of
                :py:class:`~dantro.base.BaseDataGroup` or a derived class
            RuntimeError: When ``raise_on_error`` was given and there was an
                error during saving.
        """
        # Cannot store groups
        if isinstance(result, BaseDataGroup):
            if not ignore_groups:
                raise NotImplementedError("Cannot currently write dantro "
                    "groups to a cache file. Sorry. Adjust the ignore_groups "
                    "argument in the file cache write options for the "
                    "transformation resulting in {}.".format(result))
            return False

        # Make sure the directory exists
        os.makedirs(self.cache_dir, exist_ok=True)

        # Prepare the file path (still lacks an extension)
        fpath = os.path.join(self.cache_dir, trf_hash)

        # Go over the saving functions and see if the type agrees. If so, use
        # that function to write the data.
        for types, sfunc in DAG_CACHE_RESULT_SAVE_FUNCS.items():
            if not isinstance(result, types):
                continue

            # else: type matched, invoke saving function
            try:
                sfunc(result, fpath, **save_kwargs)

            except Exception as exc:
                msg = ("Failed saving transformation cache file for result of "
                       "type {} using storage function for type(s) {}. Value "
                       "of result: {}. Additional keyword arguments: {}"
                       "".format(type(result), types, result, save_kwargs))
                if raise_on_error:
                    raise RuntimeError(msg) from exc

                log.warning("%s. %s: %s", msg, exc.__class__.__name__, exc)
                # ... will attempt pickling below
            
            else:
                # Success
                return True

        # Reached the end of the loop without returning -> not saved yet
        if not attempt_pickling:
            return False
        
        # Try to pickle it
        try:
            with open(fpath + ".pkl", mode='wb') as pkl_file:
                pkl.dump(result, pkl_file,
                         **(pkl_kwargs if pkl_kwargs else {}))
        
        except Exception as exc:
            msg = ("Failed saving transformation cache file. Cannot pickle "
                   "result object of type {} and with value {}. Consider "
                   "deactivating file caching or pickling for this "
                   "transformation."
                   "".format(type(result), result))
            if raise_on_error:
                raise RuntimeError(msg) from exc
                
            log.warning("%s. %s: %s", msg, exc.__class__.__name__, exc)
            return False
        
        # else: Success
        return True




"""This module implements the base PlotCreator class.

Classes derived from this class create plots for single files.

The interface is defined as an abstract base class and partly implemented by
the BasePlotCreator (which still remains abstract).
"""

import os
import copy
import logging
from typing import Union, Tuple, Sequence

from paramspace import ParamSpace

from ..abc import AbstractPlotCreator
from ..tools import recursive_update
from ..data_mngr import DataManager
from ..dag import TransformationDAG

# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------

class BasePlotCreator(AbstractPlotCreator):
    """The base class for PlotCreators
    
    Note that the ``plot`` method remains abstract, thus this class needs to be
    subclassed and the method implemented!
    
    Attributes:
        DAG_INVOKE_IN_BASE (bool): Whether the DAG should be created and
            computed here (in the base class). If False, the base class does
            nothing to create or compute it and the derived classes have to
            take care of it on their own.
        DAG_SUPPORTED (bool): Whether the data selection and transformation
            interface is supported by this PlotCreator. If False, the related
            methods will not be called.
        DEFAULT_EXT (str): The class variable to use for default extension.
        default_ext (str): The property-managed actual value for the default
            extension to use. This value is needed by the PlotManager in order
            to generate an out_path. It can be changed during runtime, but
            not by passing arguments to __call__, as at that point the out_path
            already needs to be fixed.
        DEFAULT_EXT_REQUIRED (bool): Whether a default extension is required
            or not. If True and the default_ext property evaluates to False,
            an error will be raised.
        EXTENSIONS (tuple): The supported extensions. If 'all', no checks for
            the extensions are performed
        OUT_PATH_EXIST_OK (bool): Whether a warning should be shown (instead
            of an error, when a plot file already exists at the specified
            output path
        POSTPONE_PATH_PREPARATION (bool): Whether to prepare paths in the base
            class's __call__ method or not. If the derived class wants to
            take care of this on their own, this should be set to True and the
            _prepare_path method, adjusted or not, should be called at another
            point of the plot execution.
    """
    EXTENSIONS = 'all'
    DEFAULT_EXT = None
    DEFAULT_EXT_REQUIRED = True
    POSTPONE_PATH_PREPARATION = False
    OUT_PATH_EXIST_OK = False
    DAG_SUPPORTED = False
    DAG_INVOKE_IN_BASE = True

    def __init__(self, name: str, *, dm: DataManager,
                 default_ext: str=None,
                 exist_ok: bool=None,
                 **plot_cfg):
        """Create a PlotCreator instance for a plot with the given ``name``.
        
        Args:
            name (str): The name of this plot
            dm (DataManager): The data manager that contains the data to plot
            default_ext (str, optional): The default extension to use; needs
                to be in ``EXTENSIONS``, if that class variable is not set to
                'all'. The value given here is needed by the PlotManager to
                build the output path.
            exist_ok (bool, optional): If True, no error will be raised when
                a plot already exists at the specified output path. If None,
                the value specified in the ``OUT_PATH_EXIST_OK`` class variable
                will be used to determine this behaviour.
            **plot_cfg: The default configuration for the plot(s) that this
                creator is supposed to create.
        
        Raises:
            ValueError: On bad `default_ext` argument
        """
        self._name = name
        self._dm = dm
        self._plot_cfg = plot_cfg
        self._exist_ok = (self.OUT_PATH_EXIST_OK if exist_ok is None
                          else exist_ok)

        # Initialize property-managed attributes
        self._logstr = None
        self._default_ext = None

        # And others via their property setters
        # Set the default extension, first from argument, then default.
        if default_ext is not None:
            self.default_ext = default_ext
        
        elif self.DEFAULT_EXT is not None:
            self.default_ext = self.DEFAULT_EXT

        # Check that it was set correctly
        if self.DEFAULT_EXT_REQUIRED and not self.default_ext:
            raise ValueError("{} requires a default extension, but neither "
                             "the argument ('{}') nor the DEFAULT_EXT class "
                             "variable ('{}') evaluated to True."
                             "".format(self.logstr, default_ext,
                                       self.DEFAULT_EXT))

    # .........................................................................
    # Properties

    @property
    def name(self) -> str:
        """Returns this creator's name"""
        return self._name

    @property
    def classname(self) -> str:
        """Returns this creator's class name"""
        return self.__class__.__name__
    
    @property
    def logstr(self) -> str:
        """Returns the classname and name of this object; a combination often
        used in logging..."""
        if not self._logstr:
            self._logstr = "{} for '{}'".format(self.classname, self.name)
        return self._logstr

    @property
    def dm(self) -> DataManager:
        """Return the DataManager"""
        return self._dm

    @property
    def plot_cfg(self) -> dict:
        """Returns a deepcopy of the plot configuration, assuring that plot
        configurations are completely independent of each other.
        """
        return copy.deepcopy(self._plot_cfg)

    @property
    def default_ext(self) -> str:
        """Returns the default extension to use for the plots"""
        return self._default_ext

    @default_ext.setter
    def default_ext(self, val: str) -> None:
        """Sets the default extension. Needs to be in EXTENSIONS"""
        if self.EXTENSIONS != 'all' and val not in self.EXTENSIONS:
            raise ValueError("Extension '{}' not supported in {}. Supported "
                             "extensions are: {}"
                             "".format(val, self.logstr, self.EXTENSIONS))

        self._default_ext = val


    # .........................................................................
    # Main API functions, required by PlotManager

    def __call__(self, *, out_path: str, **update_plot_cfg):
        """Perform the plot, updating the configuration passed to __init__
        with the given values and then calling _plot.
        
        Args:
            out_path (str): The full output path to store the plot at
            **update_plot_cfg: Keys with which to update the default plot
                configuration
        
        Returns:
            The return value of the
            :py:meth:`~dantro.plot_creators.pcr_base.BasePlotCreator.plot`
            method, which is an abstract method in
            :py:class:`~dantro.plot_creators.pcr_base.BasePlotCreator`.
        """
        # TODO add logging messages

        # Get (a deep copy of) the initial plot config
        cfg = self.plot_cfg

        # Check if a recursive update needs to take place
        if update_plot_cfg:
            cfg = recursive_update(cfg, copy.deepcopy(update_plot_cfg))

        # Find out if it's ok if out_path already exists, then prepare the path
        exist_ok = self._exist_ok
        if 'exist_ok' in cfg:
            exist_ok = cfg.pop('exist_ok')

        if not self.POSTPONE_PATH_PREPARATION:
            self._prepare_path(out_path, exist_ok=exist_ok)

        # Perform data selection and transformation, if the plot creator class
        # supports it. 
        # Even if the creator supports it, it might be disabled in the config;
        # in that case, the method below behaves like a passthrough of the cfg,
        # filtering out all transformation-related arguments.
        if self.DAG_SUPPORTED and self.DAG_INVOKE_IN_BASE:
            use_dag = cfg.pop('use_dag', None)
            _, cfg = self._perform_data_selection(use_dag=use_dag,
                                                  plot_kwargs=cfg)

        # Now call the plottig function with these arguments
        return self.plot(out_path=out_path, **cfg)

    def get_ext(self) -> str:
        """Returns the extension to use for the upcoming plot by checking
        the supported extensions and can be subclassed to have different
        behaviour.
        """
        return self.default_ext

    def prepare_cfg(self, *, plot_cfg: dict, pspace: Union[ParamSpace, dict]
                    ) -> Tuple[dict, ParamSpace]:
        """Prepares the plot configuration for the PlotManager.

        This function is called by the plot manager before the first plot
        is created.
        
        The base implementation just passes the given arguments through.
        However, it can be re-implemented by derived classes to change the
        behaviour of the plot manager, e.g. by converting a plot configuration
        to a parameter space.
        """
        return plot_cfg, pspace

    def can_plot(self, creator_name: str, **plot_cfg) -> bool:
        """Whether this plot creator is able to make a plot for the given plot
        configuration. By default, this always returns false.
        
        Args:
            creator_name (str): The name for this creator used within the
                PlotManager.
            **plot_cfg: The plot configuration with which to decide this.
        
        Returns:
            bool: Whether this creator can be used for the given plot config
        """
        return False

    # .........................................................................
    # Helpers

    def _prepare_path(self, out_path: str, *, exist_ok: bool) -> None:
        """Prepares the output path, creating directories if needed, then
        returning the full absolute path.
        
        This is called from __call__ and is meant to postpone directory
        creation as far as possible.
        
        Args:
            out_path (str): The absolute output path to start with
            exist_ok (bool): If True, will emit a warning instead of an error
        
        Raises:
            FileExistsError: Raised on already existing out path and exist_ok
                being False.
        """
        # Check that the file path does not already exist:
        if os.path.exists(out_path):
            msg = ("There already exists a file at the specified output path "
                   "for {}:\n  {}".format(self.logstr, out_path))
            if not exist_ok:
                raise FileExistsError(msg)
            log.warning(msg)

        # Ensure that all necessary directories exist
        os.makedirs(os.path.dirname(out_path), exist_ok=True)

        # Nothing more to do here, at least in the base class
    
    # .........................................................................
    # Data selection interface, using TransformationDAG

    def _perform_data_selection(self, *, use_dag: bool=None, plot_kwargs: dict,
                                **shared_kwargs) -> Tuple[bool, dict]:
        """If this plot creator supports data selection and transformation, it
        is carried out in this method.
        
        This method uses a number of other private methods to carry out the
        setup of the DAG, computing it and combining its results with the
        remaining plot configuration. Those methods have access to a subset of
        the whole configuration, thus allowing to parse the parameters that
        they need.
        
        .. note::
        
            For specializing the behaviour of the data selection and transform,
            it is best to specialize *NOT* this method, but the more granular
            DAG-related private methods.
        
        Args:
            use_dag (bool, optional): The main toggle for whether the DAG
                should be used or not. This is passed as default value to
                another method, which takes the final decision on whether the
                DAG is used or not. If None, will NOT use the DAG.
            plot_kwargs (dict): The plot configuration
            **shared_kwargs: Shared keyword arguments that are passed through
                to the helper methods ``_use_dag`` and ``_get_dag_params``
        
        Returns:
            Tuple[bool, dict]: Whether data selection was used and the plot
                configuration that can be passed on to the main ``plot``
                method.
        """
        # Determine whether the DAG framework should be used or not
        if not self._use_dag(use_dag=use_dag, plot_kwargs=plot_kwargs,
                             **shared_kwargs):
            # Only return the plot configuration, without DAG-related keys
            return False, plot_kwargs

        # Extract DAG-related parameters from the plot configuration. These are
        # not available in the plotting function.
        dag_params, plot_kwargs = self._get_dag_params(**plot_kwargs,
                                                       **shared_kwargs)

        # else: DAG should be used -> Create and compute it.
        dag = self._create_dag(**dag_params['init'])
        dag_results = self._compute_dag(dag, **dag_params['compute'])

        # Prepare the parameters passed back to __call__ and on to self.plot
        kws = self._combine_dag_results_and_plot_cfg(dag=dag,
                                                     dag_results=dag_results,
                                                     dag_params=dag_params,
                                                     plot_kwargs=plot_kwargs)
        return True, kws

    def _get_dag_params(self, *,
                        select: dict=None, transform: Sequence[dict]=None,
                        compute_only: Sequence[str]=None,
                        dag_options: dict=None,
                        **plot_kwargs) -> Tuple[dict, dict]:
        """Filters out parameters needed for DAG initialization and compute
        
        Args:
            select (dict, optional): DAG selection
            transform (Sequence[dict], optional): DAG transformation
            compute_only (Sequence[str], optional): DAG tags to be computed
            dag_options (dict, optional): Other DAG options for initialization
            **plot_kwargs: The full plot configuration
        
        Returns:
            Tuple[dict, dict]: Tuple of DAG parameters and plot kwargs
        """
        # Top-level arguments
        init_kwargs = dict(select=select, transform=transform)
        compute_kwargs = dict(compute_only=compute_only)

        # Options. Only add those, if available
        if dag_options:
            init_kwargs = dict(**init_kwargs, **dag_options)

        return dict(init=init_kwargs, compute=compute_kwargs), plot_kwargs

    def _use_dag(self, *, use_dag: bool, plot_kwargs: dict, **_kws) -> bool:
        """Whether the data transformation framework should be used.
        
        Args:
            use_dag (bool): The value from the plot configuration
            plot_kwargs (dict): The plot configuration
            **_kws: Any further kwargs that can be used to assess whether the
                DAG should be used or not. Ignored here.
        
        Returns:
            bool: Whether the DAG should be used or not
        """
        return (use_dag if use_dag is not None else False)

    def _create_dag(self, **dag_params) -> TransformationDAG:
        """Creates the actual DAG object"""
        return TransformationDAG(dm=self.dm, **dag_params)

    def _compute_dag(self, dag: TransformationDAG, **compute_kwargs) -> dict:
        """Compute the dag results"""
        return dag.compute(**compute_kwargs)

    def _combine_dag_results_and_plot_cfg(self, *, dag: TransformationDAG,
                                          dag_results: dict, dag_params: dict,
                                          plot_kwargs: dict) -> dict:
        """Combines DAG reuslts and plot configuration into one dict. The
        returned dict is then passed along to the ``plot`` method.

        The base class method ditches the ``dag_params`` and only retains the
        results, the DAG object itself, and (of course) all the remaining plot
        configuration.

        .. note::

            When subclassing, this is the method to overwrite or extend in
            order to affect which data gets passed on.
        """
        # Build a dict, delibaretly excluding the dag_params
        return dict(dag=dag, dag_results=dag_results, **plot_kwargs)



"""dantro lets you load and manipulate hierarchically organized data"""

# Package version
__version__ = '0.12.0a5'

# Configure the logging module for the whole package here by importing the
# dantro-specific logging module
from .logging import getLogger
log = getLogger(__name__)

# Make manager classes available
from .data_mngr import DataManager
from .plot_mngr import PlotManager



"""This module sets up matplotlib to be used in the other modules."""

import logging

import matplotlib as mpl
import matplotlib.pyplot as plt


# Local constants
log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
# Can define helper functions here



"""This module implements general mixin classes for containers and groups"""

import logging

# Local variables
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class ForwardAttrsMixin:
    """This Mixin class forwards all calls to unavailable attributes to a
    certain other attribute, specified by ``FORWARD_ATTR_TO`` class variable.
    """
    # The name of the existing attribute to forward to. For None, this behaves
    # as if no forwarding would occur, i.e. as if __getattr__ was not called.
    FORWARD_ATTR_TO = None

    # If set, the only attributes to be forwarded
    FORWARD_ATTR_ONLY = None

    # Attributes to _not_ forward. Evaluated after ``FORWARD_ATTR_ONLY``
    FORWARD_ATTR_EXCLUDE = ()

    def __getattr__(self, attr_name: str):
        """Forward attributes that were not available in this class to some
        other attribute of the group or container.
        
        Args:
            attr_name (str): The name of the attribute that was tried to be
                accessed but was not available in ``self``.
        
        Returns:
            The attribute ``attr_name`` of getattr(self, self.FORWARD_ATTR_TO)
        """
        if self.FORWARD_ATTR_TO is None:
            raise AttributeError(attr_name)

        if (    self.FORWARD_ATTR_ONLY is not None
            and attr_name not in self.FORWARD_ATTR_ONLY):
            raise AttributeError(attr_name)

        if attr_name in self.FORWARD_ATTR_EXCLUDE:
            raise AttributeError(attr_name)

        # Invoke the pre-hook
        self._forward_attr_pre_hook(attr_name)

        # Get the attribute
        a = getattr(self._forward_attr_get_forwarding_target(), attr_name)

        # Pass it through the post-hook
        return self._forward_attr_post_hook(a)

    def _forward_attr_pre_hook(self, attr_name: str=None):
        """Invoked before attribute forwarding occurs"""
        pass
    
    def _forward_attr_get_forwarding_target(self):
        """Get the object that the attribute call is to be forwarded to"""
        return getattr(self, self.FORWARD_ATTR_TO)
    
    def _forward_attr_post_hook(self, attr):
        """Invoked before attribute forwarding occurs"""
        return attr

class ForwardAttrsToDataMixin(ForwardAttrsMixin):
    """This Mixin class forwards all calls to unavailable attributes to the
    ``data`` attribute (a property) and thus allows to replace all behaviour
    that is not implemented in the group or container with that of the stored
    data.
    """

    FORWARD_ATTR_TO = 'data'



"""In this module, BaseDataContainer specializations that make use of features
from the paramspace package are implemented.
"""

import copy
import logging
from typing import Union, List, Dict, Callable, Sequence

import numpy as np
import numpy.ma
import xarray as xr

from paramspace import ParamSpace

from .ordered import OrderedDataGroup, IndexedDataGroup
from ..utils.data_ops import multi_concat as _multi_concat
from ..base import PATH_JOIN_CHAR
from ..containers import NumpyDataContainer, XrDataContainer
from ..mixins import PaddedIntegerItemAccessMixin

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class ParamSpaceStateGroup(OrderedDataGroup):
    """A ParamSpaceStateGroup is a member group of the
    :py:class:`~dantro.groups.pspgrp.ParamSpaceGroup`.

    While its *own* name need be interpretable as a positive integer (enforced
    in the enclosing :py:class:`~dantro.groups.pspgrp.ParamSpaceGroup` but also
    here), it can *hold* members with any name.
    """
    _NEW_GROUP_CLS = OrderedDataGroup

    def _check_name(self, name: str) -> None:
        """Called by __init__ and overwritten here to check the name."""
        # Assert that the name is valid, i.e. convertible to an integer
        try:
            int(name)

        except ValueError as err:
            raise ValueError("Only names that are representible as integers "
                             "are possible for the name of {}, got '{}'!"
                             "".format(self.classname, name)
                             ) from err

        # ... and not negative
        if int(name) < 0:
            raise ValueError("Name for {} needs to be positive when converted "
                             "to integer, was: {}"
                             "".format(self.classname, name))

        # Still ask the parent method for its opinion on this matter
        super()._check_name(name)


class ParamSpaceGroup(PaddedIntegerItemAccessMixin, IndexedDataGroup):
    """The ParamSpaceGroup is associated with a ParamSpace object and the
    loaded results of an iteration over this parameter space.

    Thus, the groups that are stored in the ParamSpaceGroup need all relate to
    a state of the parameter space, identified by a zero-padded string name.
    In fact, this group allows no other kinds of groups stored inside.

    To make access to a specific state easier, it allows accessing a state by
    its state number as integer.
    """

    # Class variables that define some of the behaviour
    # Define which .attrs entry to return from the `pspace` property
    _PSPGRP_PSPACE_ATTR_NAME = 'pspace'

    # A transformation callable that can be used during data selection
    _PSPGRP_TRANSFORMATOR = None

    # Define the class to use for the direct members of this group
    _NEW_GROUP_CLS = ParamSpaceStateGroup

    # Define allowed container types
    _ALLOWED_CONT_TYPES = (ParamSpaceStateGroup,)

    # .........................................................................

    def __init__(self, *, name: str, pspace: ParamSpace=None,
                 containers: list=None, **kwargs):
        """Initialize a OrderedDataGroup from the list of given containers.
        
        Args:
            name (str): The name of this group.
            pspace (ParamSpace, optional): Can already pass a ParamSpace object
                here.
            containers (list, optional): A list of containers to add, which
                need to be ParamSpaceStateGroup objects.
            **kwargs: Further initialisation kwargs, e.g. `attrs` ...
        """
        # Initialize with parent method, which will call .add(*containers)
        super().__init__(name=name, containers=containers, **kwargs)

        # If given, associate the parameter space object
        if pspace is not None:
            self.pspace = pspace

    # Properties ..............................................................

    @property
    def pspace(self) -> Union[ParamSpace, None]:
        """Reads the entry named _PSPGRP_PSPACE_ATTR_NAME in .attrs and
        returns a ParamSpace object, if available there.
        
        Returns:
            Union[ParamSpace, None]: The associated parameter space, or None,
                if there is none associated yet.
        """
        return self.attrs.get(self._PSPGRP_PSPACE_ATTR_NAME, None)

    @pspace.setter
    def pspace(self, val: ParamSpace):
        """If not already set, sets the entry in the attributes that is
        accessed by the .pspace property
        """
        if self.pspace is not None:
            raise RuntimeError("The attribute for the parameter space of this "
                               "{} was already set, cannot set it again!"
                               "".format(self.logstr))

        
        elif not isinstance(val, ParamSpace):
            raise TypeError("The attribute for the parameter space of {} "
                            "needs to be a ParamSpace-derived object, was {}!"
                            "".format(self.logstr, type(val)))

        # Checked it, now set it
        self.attrs[self._PSPGRP_PSPACE_ATTR_NAME] = val

        log.debug("Associated %s with %s", val, self.logstr)

    @property
    def only_default_data_present(self) -> bool:
        """Returns true if only data for the default point in parameter space
        is available in this group.
        """
        return (len(self) == 1) and (0 in self)


    # Data access .............................................................

    def select(self, *,
               field: Union[str, List[str]]=None,
               fields: Dict[str, List[str]]=None,
               subspace: dict=None,
               method: str='concat',
               idx_as_label: bool=False,
               base_path: str=None,
               **kwargs) -> xr.Dataset:
        """Selects a multi-dimensional slab of this ParamSpaceGroup and the
        specified fields and returns them bundled into an ``xarray.Dataset``
        with labelled dimensions and coordinates.
        
        Args:
            field (Union[str, List[str]], optional): The field of data to
                select. Should be path or a list of strings that points to an
                entry in the data tree. To select multiple fields, do not pass
                this argument but use the `fields` argument.
            fields (Dict[str, List[str]], optional): A dict specifying the
                fields that are to be loaded into the dataset. Keys will be
                the names of the resulting variables, while values should
                specify the path to the field in the data tree. Thus, they can
                be strings, lists of strings or dicts with the `path` key
                present. In the latter case, a dtype can be specified via the
                `dtype` key in the dict.
            subspace (dict, optional): Selector for a subspace of the
                parameter space. Adheres to the ParamSpace.activate_subspace
                signature.
            method (str, optional): How to combine the selected datasets.
        
                    - ``concat``: concatenate sequentially along all parameter
                      space dimensions. This can preserve the data type but
                      it does not work if one data point is missing.
                    - ``merge``: merge always works, even if data points are
                      missing, but will convert all dtypes to float.
        
            idx_as_label (bool, optional): If true, adds the trivial indices
                as labels for those dimensions where coordinate labels were not
                extractable from the loaded field. This allows merging for data
                with different extends in an unlabelled dimension.
            base_path (str, optional): If given, ``path`` specifications for
                each field can be seen as relative to this path
            **kwargs: Passed along either to xr.concat or xr.merge, depending
                on the ``method`` argument.
        
        Raises:
            KeyError: Description
            ValueError: Raised in multiple scenarios:
        
                - If no ParamSpace was associated with this group
                - For wrong argument values
                - If the data to select cannot be extracted with the given
                  argument values
                - Exceptions passed on from xarray
        
        Returns:
            xr.Dataset: The selected hyperslab of the parameter space, holding
                the desired fields.
        """

        def parse_fields(*, field, fields) -> dict:
            """Parses the field and fields arguments  into a uniform dict

            Return value is a dict of the following structure:
            <var_name_1>:
              path: <list of strings>
              dtype: <str, optional>
              dims: <list of strings, optional>
              ... further
            <var_name_2>:
              ...

            TODO Change such that its using strings for paths, not sequences.
            """

            if field is not None and fields is not None:
                raise ValueError("Can only specify either of the arguments "
                                 "`field` or `fields`, got both!")

            elif field is None and fields is None:
                raise ValueError("Need to specify one of the arguments "
                                 "`field` or `fields`, got neither of them!")

            elif field is not None:
                # Generate a dict from the single field argument and put it
                # into a fields dict such that it can be processed like the
                # rest ...

                # Need to find a name from the path
                if isinstance(field, str):
                    path = field.split(PATH_JOIN_CHAR)
                    kwargs = {}

                elif isinstance(field, dict):
                    path = field['path']
                    kwargs = {k:v for k, v in field.items() if k != "path"}
                    # Not using .pop here in order to not change the dict.

                    if isinstance(path, str):
                        path = path.split(PATH_JOIN_CHAR)
                    
                else:
                    path = list(field)
                    kwargs = {}

                # Create the fields dict, carrying over all other arguments
                fields = dict()
                fields[path[-1]] = dict(path=path, **kwargs)

            # The fields variable is now available
            # Make sure it is of right type
            if not isinstance(fields, dict):
                raise TypeError("Argument `fields` needs to be a dict, "
                                "but was {}!".format(type(fields)))

            
            # Ensure values of the dict are dicts of the proper structre
            for name, field in fields.items():
                if isinstance(field, str):
                    fields[name] = dict(path=field.split(PATH_JOIN_CHAR))

                elif not isinstance(field, dict):
                    # Assume this is a sequence, but better make sure ...
                    fields[name] = dict(path=list(field))

                # else: Already a dict; nothing to do. Parameters carried over.

            return fields

        def get_state_grp(state_no: int) -> ParamSpaceStateGroup:
            """Returns the group corresponding to the given state"""
            try:
                return self[state_no]

            except (KeyError, ValueError) as err:
                # TODO use custom exception class, e.g. from DataManager?
                raise ValueError("No state {} available in {}! Make sure the "
                                 "data was fully loaded."
                                 "".format(state_no, self.logstr)) from err

        def get_var(state_grp: ParamSpaceStateGroup, *,
                    path: List[str],
                    base_path: List[str]=None,
                    dtype: str=None,
                    dims: List[str]=None,
                    transform: Sequence[dict]=None,
                    **transform_kwargs) -> Union[xr.Variable, xr.DataArray]:
            """Extracts the field specified by the given path and returns it as
            either an xr.Variable or (for supported containers) directly as an
            xr.DataArray.
            
            We are using xr.Variables as defaults here, as they provide higher
            performance than xr.DataArrays; the latter have to be frequently
            unpacked and restructured in the merge operations.
            
            Args:
                state_grp (ParamSpaceStateGroup): The group to search `path` in
                path (List[str]): The path to a data container.
                base_path (List[str], optional): Will be prepended to the given
                    path, if given.
                dtype (str, optional): The desired dtype for the data.
                dims (List[str], optional): A list of dimension names for the
                    extracted data. If not given, will name them manually as
                    dim_0, dim_1, ...
                transform (Sequence[dict], optional): Optional transform
                    arguments; passed on to transformator as *args.
                **transform_kwargs: Passed on to the transformator as **kwargs.
            
            Returns:
                Union[xr.Variable, xr.DataArray]: The extracted data, which
                    can be either a data array (if the path led to an
                    xarray-interface supporting container) or an xr.Variable
                    (if not).
            
            Raises:
                ValueError: Description
            """
            def convert_dtype(data, dtype, *, path):
                """Change the dtype of the data, if it does not match the
                specified one.
                """
                if data.dtype == dtype:
                    return data
                log.debug("Converting data from '%s' with dtype %s to %s ...",
                          PATH_JOIN_CHAR.join(path), data.dtype, dtype)
                return data.astype(dtype)

            # Prepare the path, ensuring to work on the list representation
            if not isinstance(path, list):
                path = path.split(PATH_JOIN_CHAR)

            if base_path:
                path = base_path + path

            # Now, get the desired container
            cont = state_grp[path]

            # Apply the transformator on the container, if arguments given
            if transform or transform_kwargs:
                if self._PSPGRP_TRANSFORMATOR is None:
                    raise ValueError("Got transform arguments or kwargs, but "
                                     "no transformator callable was defined "
                                     "as class variable!")

                # Invoke the transformator on the container
                cont = self._PSPGRP_TRANSFORMATOR(cont, *transform,
                                                  **transform_kwargs)
            
            # Shortcut: specialised containers might already supply all the
            # information, including coordinates. In that case, return it as
            # a data array.
            if isinstance(cont, XrDataContainer):
                # Will return the underlying data. See if some dtype change
                # or dimension name relabelling was specified
                darr = cont.data

                if dtype is not None:
                    darr = convert_dtype(darr, dtype, path=path)

                if dims is not None:
                    darr = darr.rename({old:new for old, new
                                        in zip(darr.dims, dims)})

                return darr

            elif isinstance(cont, (xr.DataArray, xr.Dataset)):
                # Actually was not a container but already the data; skip below
                return cont

            # If this was not the case, xr.Variable will have to be constructed
            # manually from the container.
            # The only pre-requisite for the data is that it is np.array-like,
            # which is always possible; worst case: scalar of dtype "object".
            data = np.array(cont.data)
            # Can now assume data to comply to np.array interface

            # Check the dtype and convert, if needed
            if dtype is not None:
                data = convert_dtype(data, dtype, path=path)

            # Get the attributes
            attrs = {k: v for k, v in cont.attrs.items()}

            # Generate dimension names, if not explicitly given.
            if dims is None:
                dims = ["dim_{}".format(i) for i in range(len(data.shape))]

            # Check whether indices are to be added (var from outer scope!)
            if not idx_as_label:
                # Can use these to construct an xr.Variable
                return xr.Variable(dims, data, attrs=attrs)

            # else: will need to be a DataArray; Variable does not hold coords
            # For each dimension, add trivial coordinates
            coords = {d: range(data.shape[i]) for i, d in enumerate(dims)}

            return xr.DataArray(data, dims=dims, coords=coords, attrs=attrs)

        def combine(*, method: str,
                    dsets: np.ndarray, psp: ParamSpace) -> xr.Dataset:
            """Tries to combine the given datasets either by concatenation or
            by merging and returns a combined xr.Dataset
            """
            # NOTE change for valid `method` value is carried out before this
            #      function is called.

            # Merging . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
            if method in ['merge']:
                log.remark("Combining datasets by merging ...")
                # TODO consider warning about dtype changes?!

                dset = xr.merge(dsets.flat)

                log.remark("Merge successful.")
                return dset

            # else: Concatenation . . . . . . . . . . . . . . . . . . . . . . .
            log.remark("Combining %d datasets by concatenation along %d "
                       "dimensions ...", dsets.size, len(dsets.shape))

            # Reduce the dsets array to one dimension by applying xr.concat
            # along each axis. The returned object contains the combined data.
            reduced = _multi_concat(dsets, dims=psp.dims.keys())
            log.remark("Concatenation successful.")

            return reduced


        # End of definition of helper functions.
        # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # Some initial checks

        if self.pspace is None:
            raise ValueError("Cannot get data from {} without having a "
                             "parameter space associated!".format(self.logstr))

        elif method not in ('concat', 'merge'):
            raise ValueError("Invalid value for argument `method`: '{}'. Can "
                             "be: 'concat' (default), 'merge'".format(method))

        # Pre-process arguments . . . . . . . . . . . . . . . . . . . . . . . .
        # From field and fields arguments, generate a fields dict, such that
        # it can be handled uniformly below.
        fields = parse_fields(field=field, fields=fields)

        # Prepare the base path
        if base_path and not isinstance(base_path, list):
            base_path = base_path.split(PATH_JOIN_CHAR)

        # Work on a copy of the parameter space and apply the subspace masks
        psp = copy.deepcopy(self.pspace)

        if subspace:
            # Need the parameter space to be of non-zero volume
            if psp.volume == 0:
                raise ValueError("Cannot select a subspace because the "
                                 "associated parameter space has no "
                                 "dimensions defined! Remove the `subspace` "
                                 "argument in the call to this method.")

            try:
                psp.activate_subspace(**subspace)

            except KeyError as err:
                raise KeyError("Could not select a subspace! {}: {}\n"
                               "Make sure your subspace selector contains "
                               "only valid dimension names and coordinates. "
                               "Available dimension names: {}"
                               "".format(type(err).__name__, err,
                                         ", ".join(psp.dims.keys()))
                               ) from err

        # Now, the data needs to be collected from each point in this subspace
        # and associated with the correct coordinate, such that the datasets
        # can later be merged and aligned by that coordinate.
        if psp.volume > 0:
            log.info("Collecting data for %d fields from %d points in "
                     "parameter space ...", len(fields), psp.volume)
        else:
            log.info("Collecting data for %d fields from a dimensionless "
                     "parameter space ...", len(fields))

        # Gather them in a multi-dimensional array
        dsets = np.zeros(psp.shape, dtype="object")
        dsets.fill(dict())  # these are ignored in xr.merge

        # Prepare the iterators
        psp_it = psp.iterator(with_info=('state_no', 'current_coords'),
                              omit_pt=True)
        arr_it = np.nditer(dsets, flags=('multi_index', 'refs_ok'))

        for (_state_no, _coords), _ in zip(psp_it, arr_it):
            # Select the corresponding state group
            try:
                _state_grp = get_state_grp(_state_no)

            except ValueError:
                if method == 'merge':
                    # In merge, this will mereley lead to a NaN ...
                    log.warning("Missing state group:  %d", _state_no)
                    continue
                # ...but for concatenation, it will result in an error.
                raise

            # Get the variables for all fields
            _vars = {k: get_var(_state_grp, **f, base_path=base_path)
                     for k, f in fields.items()}

            # Construct a dataset from that ...
            _dset = xr.Dataset(_vars)

            # ... and expand its dimensions to accomodate the point in pspace
            _dset = _dset.expand_dims({k: [v] for k, v in _coords.items()})

            # Store it in the array of datasets
            dsets[arr_it.multi_index] = _dset

        # All data points collected now.
        log.info("Data collected.")

        # Finally, combine all the datasets together into a dataset with
        # potentially non-homogeneous data type. This will have at least the
        # dimensions given by the parameter space aligned, but there could
        # be potentially more dimensions.

        try:
            dset = combine(method=method, dsets=dsets, psp=psp)

        except ValueError as err:
            raise ValueError("Combination of datasets failed; see below. This "
                             "is probably due to a failure of alignment, "
                             "which can be resolved by adding trivial "
                             "coordinates (i.e.: the indices) to unlabelled "
                             "dimensions by setting the `idx_as_label` "
                             "argument to True.") from err

        log.info("Data selected.")
        log.note("Available data variables:      %s",
                 ", ".join(dset.data_vars))
        log.note("Dataset dimensions and sizes:  %s",
                 ", ".join("{}: {}".format(*kv) for kv in dset.sizes.items()))
        return dset



"""This module implements the base classes of dantro, based on the abstract
classes.

The base classes are classes that combine features of the abstract classes. For
example, the data group gains attribute functionality by being a combination
of the AbstractDataGroup and the BaseDataContainer.
In turn, the BaseDataContainer uses the BaseDataAttrs class as an attribute and
thereby extends the AbstractDataContainer class.

NOTE: These classes are not meant to be instantiated but used as a basis to
implement more specialized BaseDataGroup- or BaseDataContainer-derived classes.
"""

import abc
import copy
import logging
import inspect
from typing import Union, List, Tuple, Callable

import dantro.abc
from .abc import PATH_JOIN_CHAR
from .mixins import AttrsMixin, SizeOfMixin, CheckDataMixin, LockDataMixin
from .mixins import CollectionMixin, ItemAccessMixin, MappingAccessMixin
from .tools import TTY_COLS

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class BaseDataProxy(dantro.abc.AbstractDataProxy):
    """The base class for data proxies.

    NOTE: This is still an abstract class and needs to be subclassed.
    """

    @abc.abstractmethod
    def __init__(self, obj):
        """Initialize a proxy object for the given object."""
        log.trace("Initialising %s for %s ...", self.classname, type(obj))

        # Specify the tags of this proxy; none by default
        self._tags = tuple()

    @property
    def tags(self) -> Tuple[str]:
        """The tags describing this proxy object"""
        return self._tags


# -----------------------------------------------------------------------------

class BaseDataAttrs(MappingAccessMixin, dantro.abc.AbstractDataAttrs):
    """A class to store attributes that belong to a data container.

    This implements a dict-like interface and serves as default attribute
    class.

    NOTE: Unlike the other base classes, this can already be instantiated. That
    is required as it is needed in BaseDataContainer where no previous
    subclassing or mixin is reasonable.
    """

    def __init__(self, attrs: dict=None, **dc_kwargs):
        """Initialize a DataAttributes object.
        
        Args:
            attrs (dict, optional): The attributes to store
            **dc_kwargs: Further kwargs to the parent DataContainer
        """
        # Make sure it is a dict; initialize empty if empty
        attrs = dict(attrs) if attrs else {}

        # Store them via the parent method.
        super().__init__(data=attrs, **dc_kwargs)
    
    # .........................................................................
    
    def as_dict(self) -> dict:
        """Returns a shallow copy of the attributes as a dict"""
        return {k:v for k, v in self.items()}

    # .........................................................................
    # Magic methods and iterators for convenient dict-like access

    def _format_info(self) -> str:
        """A __format__ helper function: returns info about these attributes"""
        return "{} attribute(s)".format(len(self))


# -----------------------------------------------------------------------------

class BaseDataContainer(AttrsMixin, SizeOfMixin,
                        dantro.abc.AbstractDataContainer):
    """The BaseDataContainer extends the abstract base class by the ability to
    hold attributes and be path-aware.
    """
    # Define which class to use for storing attributes
    _ATTRS_CLS = BaseDataAttrs

    def __init__(self, *, name: str, data, attrs=None):
        """Initialize a BaseDataContainer, which can store data and attributes.
        
        Args:
            name (str): The name of this data container
            data: The data to store in this container
            attrs (None, optional): A mapping that is stored as attributes
        """
        # Initialize via parent, then additionally store attributes
        super().__init__(name=name, data=data)
        self.attrs = attrs

    def _format_info(self) -> str:
        """A __format__ helper function: returns info about the content of this
        data container.
        """
        return ("{} attribute{}"
                "".format(len(self.attrs),
                          "s" if len(self.attrs) != 1 else ""))
        

# -----------------------------------------------------------------------------

class BaseDataGroup(LockDataMixin, SizeOfMixin, AttrsMixin,
                    dantro.abc.AbstractDataGroup):
    """The BaseDataGroup serves as base group for all data groups.

    It implements all functionality expected of a group, which is much more
    than what is expected of a general container.
    """
    # Define which class to use for storing attributes
    _ATTRS_CLS = BaseDataAttrs
    
    # The mapping type that is used to store the members of this group.
    _STORAGE_CLS = dict

    # Define which class to use in the new_group method. If None, the type of
    # this instance is used
    _NEW_GROUP_CLS = None
    
    # Define which class to use in the new_container method. If None, the type
    # needs to be specified explicitly in the call to new_container
    _NEW_CONTAINER_CLS = None

    # Define the types that are allowed to be stored in this group. If None,
    # the dantro base classes are allowed
    _ALLOWED_CONT_TYPES = None

    # Condensed tree representation: maximum level
    _COND_TREE_MAX_LEVEL = 10

    # Condensed tree representation: threshold parameter
    _COND_TREE_CONDENSE_THRESH = 10

    # .........................................................................

    def __init__(self, *, name: str, containers: list=None, attrs=None):
        """Initialize a BaseDataGroup, which can store other containers and
        attributes.
        
        Args:
            name (str): The name of this data container
            containers (list, optional): The containers that are to be stored
                as members of this group. If given, these are added one by one
                using the `.add` method.
            attrs (None, optional): A mapping that is stored as attributes
        """
        # Prepare the storage class that is used to store the members
        data = self._STORAGE_CLS()

        # Initialize via parent and store attributes
        super().__init__(name=name, data=data)
        self.attrs = attrs

        # Now add the member containers
        if containers is not None:
            self.add(*containers)

    # .........................................................................
    # Item access

    def __getitem__(self, key: Union[str, List[str]]):
        """Returns the container in this group with the given name.
        
        Args:
            key (Union[str, List[str]]): The object to retrieve.
                If this is a path, will recurse down until at the end.
        
        Returns:
            The object at `key`
        
        Raises:
            KeyError: If no such key can be found
        """
        if not isinstance(key, list):
            # Assuming this is a string, i.e.: the only other type allowed
            key = key.split(PATH_JOIN_CHAR)

        # Can be sure that this is a list now
        try:
            # If there is more than one entry, need to call this recursively
            if len(key) > 1:
                return self._data[key[0]][key[1:]]
            # else: end of recursion
            return self._data[key[0]]

        except (KeyError, IndexError) as err:
            raise KeyError("No key or key sequence '{}' in {}! "
                           "Available keys at top level: {}"
                           "".format(key, self.logstr,
                                     ", ".join([k for k in self.keys()]))
                           ) from err

    def __setitem__(self, key: Union[str, List[str]],
                    val: BaseDataContainer) -> None:
        """This method is used to allow access to the content of containers of
        this group. For adding an element to this group, use the `add` method!
        
        Args:
            key (Union[str, List[str]]): The key to which to set the value.
                If this is a path, will recurse down to the lowest level.
                Note that all intermediate keys need to be present.
            val (BaseDataContainer): The value to set
        
        Returns:
            None
        
        Raises:
            ValueError: If trying to add an element to this group, which should
                be done via the `add` method.
        
        """
        if not isinstance(key, list):
            key = key.split(PATH_JOIN_CHAR)

        # Depending on length of the key sequence, start recursion or not
        if len(key) > 1:
            self._data[key[0]][key[1:]] = val
            return
        
        # else: end of recursion, i.e. the path led to an item of this group
        # This operation is not allowed, as the add method should be used
        # That method takes care that the name this element is registered with
        # is equal to that of the registered object
        raise ValueError("{} cannot carry out __setitem__ operation for the "
                         "given key '{}'. Note that to add a group or "
                         "container to the group, the `add` method should "
                         "be used.".format(self.logstr, key))

    def __delitem__(self, key: str) -> None:
        """Deletes an item from the group"""

        if not isinstance(key, list):
            # Assuming this is a string ...
            key = key.split(PATH_JOIN_CHAR)

        # Can be sure that this is a list now
        # If there is more than one entry, need to call this recursively
        if len(key) > 1:
            # Continue recursion
            del self._data[key[0]][key[1:]]
            return

        # else: end of recursion: delete and unlink this container
        # ... if it is not locked
        self.raise_if_locked()
        
        cont = self._data[key[0]]
        del self._data[key[0]]

        self._unlink_child(cont)

    def add(self, *conts, overwrite: bool=False):
        """Add the given containers to this group."""
        for cont in conts:
            self._add_container(cont, overwrite=overwrite)

        log.trace("Added %d container(s) to %s.", len(conts), self.logstr)

    def _add_container(self, cont, *, overwrite: bool):
        """Private helper method to add a container to this group."""
        # Data may not be locked
        self.raise_if_locked()

        # Check the allowed types
        if (self._ALLOWED_CONT_TYPES is None
            and not isinstance(cont, (BaseDataGroup, BaseDataContainer))):
            raise TypeError("Can only add BaseDataGroup- or "
                            "BaseDataContainer-derived objects to {}, got {}!"
                            "".format(self.logstr, type(cont)))

        elif (self._ALLOWED_CONT_TYPES is not None
              and not isinstance(cont, self._ALLOWED_CONT_TYPES)):
            raise TypeError("Can only add objects derived from the following "
                            "classes: {}. Got: {}"
                            "".format(self._ALLOWED_CONT_TYPES, type(cont)))

        # else: is of correct type
        # Check if one like this already exists
        old_cont = None
        if cont.name in self:
            if not overwrite:
                raise ValueError("{} already has a member with name '{}', "
                                 "cannot add {}."
                                 "".format(self.logstr, cont.name, cont))
            log.debug("A member '%s' of %s already exists and will be "
                      "overwritten ...", cont.name, self.logstr)
            old_cont = self[cont.name]

        # Allow for subclasses to perform further custom checks on the
        # container object before adding it
        self._check_cont(cont)

        # Write to data, assuring that the name matches that of the container
        self._add_container_to_data(cont)
        self._add_container_callback(cont)
        
        # Re-link the containers
        self._link_child(new_child=cont, old_child=old_cont)

    def _check_cont(self, cont) -> None:
        """Can be used by a subclass to check a container before adding it to
        this group. Is called by _add_container before checking whether the
        object exists or not.
        
        This is not expected to return, but can raise errors, if something
        did not work out as expected.
        
        Args:
            cont: The container to check
        """
        pass

    def _add_container_to_data(self, cont) -> None:
        """Performs the operation of adding the container to the _data. This
        can be used by subclasses to make more elaborate things while adding
        data, e.g. specify ordering ...

        NOTE This method should NEVER be called on its own, but only via the
             _add_container method, which takes care of properly linking the
             container that is to be added.

        NOTE After adding, the container need be reachable under its .name!
        
        Args:
            cont: The container to add
        """
        # Just add it via _data.__setitem__, using the container's name
        self._data[cont.name] = cont

    def _add_container_callback(self, cont) -> None:
        """Called after a container was added."""
        pass

    def new_container(self, path: Union[str, list], *,
                      Cls: type=None, **kwargs):
        """Creates a new container of class `Cls` and adds it at the given path
        relative to this group.
        
        Args:
            path (Union[str, list]): Where to add the container. Note that the
                intermediates of this path need to already exist.
            Cls (type, optional): The class of the container to add. If None,
                the _NEW_CONTAINER_CLS class variable's value is used; if not
                given, this will raise a ValueError.
            **kwargs: kwargs to pass on to Cls.__init__
        
        Returns:
            Cls: the created container
        
        Raises:
            KeyError: When intermediate groups to `path` are missing
            TypeError: When the given Cls is invalid
        """
        # Resolve the Cls argument, if possible from the class variable
        if Cls is None:
            if self._NEW_CONTAINER_CLS is None:
                raise ValueError("Got neither argument `Cls` nor class "
                                 "variable _NEW_CONTAINER_CLS, at least one "
                                 "of which is needed to determine the type "
                                 "of the new container!")

            Cls = self._NEW_CONTAINER_CLS

        # Check the class to create the container with
        if not inspect.isclass(Cls):
            raise TypeError("Argument `Cls` needs to be a class, but "
                            "was of type {} with value '{}'."
                            "".format(type(Cls), Cls))

        elif not issubclass(Cls, (BaseDataContainer, BaseDataGroup)):
            raise TypeError("Argument `Cls` needs to be a subclass of "
                            "BaseDataContainer or BaseDataGroup, was '{}'."
                            "".format(Cls))
        # Class is checked now

        # Make sure the path is a list
        if not isinstance(path, list):
            path = path.split(PATH_JOIN_CHAR)

        # Check whether recursion ends here, i.e.: the path ends here
        if len(path) == 1:
            # Yes, and it's a string: create container, add, return
            cont = Cls(name=path[0], **kwargs)
            self.add(cont)
            return cont

        # Recursive branch: need to split off the front section and continue
        grp_name, new_path = path[0], path[1:]

        try:
            return self[grp_name].new_container(new_path, Cls=Cls, **kwargs)

        except KeyError as err:
            raise KeyError("Could not create {} at '{}'! Check that all "
                           "intermediate groups have already been created. "
                           "Entries available in {}: {}"
                           "".format(Cls.__name__,
                                     PATH_JOIN_CHAR.join(path), self.logstr,
                                     ", ".join([k for k in self.keys()]))
                           ) from err

    def new_group(self, path: Union[str, list], *, Cls: type=None, **kwargs):
        """Creates a new group at the given path.
        
        Args:
            path (Union[str, list]): The path to create the group at. Note
                that the whole intermediate path needs to already exist.
            Cls (type, optional): If given, use this type to create the
                group. If not given, uses the class specified in the
                _NEW_GROUP_CLS class variable or, as last resort, the type of
                this instance.
            **kwargs: Passed on to Cls.__init__
        
        Returns:
            Cls: the created group
        
        Raises:
            TypeError: For the given class not being derived from BaseDataGroup
        """
        # If no Cls is given, use this instance's type
        if Cls is None:
            Cls = self._NEW_GROUP_CLS if self._NEW_GROUP_CLS else type(self)

        # Need to catch the case where a non-group class was given
        if inspect.isclass(Cls) and not issubclass(Cls, BaseDataGroup):
            raise TypeError("Argument `Cls` needs to be a subclass of "
                            "BaseDataGroup, was '{}'.".format(Cls))

        # Use container method to create the entry. Recursion happens there.
        return self.new_container(path, Cls=Cls, **kwargs)

    def recursive_update(self, other):
        """Recursively updates the contents of this data group with the entries
        of the given data group
        
        NOTE This will create shallow copies of those elements in `other` that
        are added to this object.
        
        Args:
            other (BaseDataGroup): The group to update with
        
        Raises:
            TypeError: If `other` was of invalid type
        """

        if not isinstance(other, BaseDataGroup):
            raise TypeError("Can only update {} with objects of classes that "
                            "are derived from BaseDataGroup. Got: {}"
                            "".format(self.logstr, type(other)))

        # Loop over the given DataGroup
        for name, obj in other.items():
            # Distinguish between the case where it is another group and where
            # it is a container
            if isinstance(obj, BaseDataGroup):
                # Already a group -> if a group with the same name is already
                # present, continue recursion. If not, just create an entry
                # and add it to this group
                if name in self:
                    # Continue recursion
                    self[name].recursive_update(obj)
                else:
                    # Can add the object, but need to detach the parent first
                    # and thus need to work on a copy
                    obj = copy.copy(obj)
                    obj.parent = None
                    self.add(obj)

            else:
                # Not a group; add a shallow copy
                obj = copy.copy(obj)
                obj.parent = None
                self.add(obj)

        log.debug("Finished recursive update of %s.", self.logstr)

    # .........................................................................
    # Linking

    # For correct child-parent linking, some helper methods
    def _link_child(self, *, new_child: BaseDataContainer,
                    old_child: BaseDataContainer=None):
        """Links the new_child to this class, unlinking the old one.

        This method should be called from any method that changes which items
        are associated with this group.
        """
        # Check that it was already associated
        if new_child not in self:
            raise ValueError("{} needs to be a child of {} _before_ it can "
                             "be linked. Use the add method to add the child "
                             "to the group.".format(new_child.logstr,
                                                    self.logstr))
        new_child.parent = self

        if old_child is not None:
            self._unlink_child(old_child)

    def _unlink_child(self, child: BaseDataContainer):
        """Unlink a child from this class.

        This method should be called from any method that removes an item from
        this group, be it through deletion or through 
        """
        if child.parent is not self:
            raise ValueError("{} was not linked to {}. Refuse to unlink."
                             "".format(child.logstr, self.logstr))
        child.parent = None

    # .........................................................................
    # Information

    def __len__(self) -> int:
        """The length of the data."""
        return len(self._data)

    def __contains__(self, cont: Union[str, BaseDataContainer]) -> bool:
        """Whether the given container is in this group or not.
        
        Args:
            cont (Union[str, BaseDataContainer]): The name of the container or 
                an object reference. 
        
        Returns:
            bool: Whether the given container is in this group.
        """
        if isinstance(cont, (BaseDataGroup, BaseDataContainer)):
            # Implement a contains-check via is-relationships
            for obj in self.values():
                if obj is cont:
                    return True 
            return False

        # Check via the string, if such a key is available
        elif isinstance(cont, str):
            key_seq = cont.split(PATH_JOIN_CHAR)

        else:
            # Need to check its type
            if not isinstance(cont, list):
                raise TypeError("Can only check content of {} against the "
                                "container name or path (as string), or the "
                                "explicit reference to the object! "
                                "Got {} with value '{}'."
                                "".format(self.logstr, type(cont), cont))

            key_seq = cont

        # key_seq is now a list of keys
        if len(key_seq) > 1:
            # More than one element.
            # If the target element exists and is a group, continue recursively
            if (key_seq[0] in self
                and isinstance(self[key_seq[0]], BaseDataGroup)):
                return key_seq[1:] in self[key_seq[0]]

            # else: does not exist or is not a group
            return False
        
        elif len(key_seq) == 1:
            return bool(key_seq[0] in self.keys())

        return False

    def _ipython_key_completions_(self) -> List[str]:
        """For ipython integration, return a list of available keys"""
        return list(self.keys())

    # .........................................................................
    # Iteration

    def __iter__(self):
        """Returns an iterator over the OrderedDict"""
        return iter(self._data)

    def keys(self):
        """Returns an iterator over the container names in this group."""
        return self._data.keys()

    def values(self):
        """Returns an iterator over the containers in this group."""
        return self._data.values()

    def items(self):
        """Returns an iterator over the (name, data container) tuple of this
        group."""
        return self._data.items()

    def get(self, key, default=None):
        """Return the container at `key`, or `default` if container with name
        `key` is not available."""
        return self._data.get(key, default)

    def setdefault(self, key, default=None):
        """This method is not supported for a data group"""
        raise NotImplementedError("setdefault is not supported by {}! Use the "
                                  "`add` method or `new_group` and "
                                  "`new_container` to add elements."
                                  "".format(self.classname))

    # .........................................................................
    # Formatting

    @property
    def tree(self) -> str:
        """Returns the default (full) tree representation of this group"""
        return self._tree_repr()
    
    @property
    def tree_condensed(self) -> str:
        """Returns the condensed tree representation of this group. Uses the
        ``_COND_TREE_*`` prefixed class attributes as parameters.
        """
        return self._tree_repr(max_level=self._COND_TREE_MAX_LEVEL,
                               condense_thresh=self._COND_TREE_CONDENSE_THRESH)

    def _format_info(self) -> str:
        """A __format__ helper function: returns an info string that is used
        to characterize this object. Does NOT include name and classname!
        """
        return ("{} member{}, {} attribute{}"
                "".format(len(self), "s" if len(self) != 1 else "",
                          len(self.attrs), "s" if len(self.attrs) != 1 else "")
                )

    def _format_tree(self) -> str:
        """Returns the default tree representation of this group by invoking
        the .tree property
        """
        return self.tree

    def _format_tree_condensed(self) -> str:
        """Returns the default tree representation of this group by invoking
        the .tree property
        """
        return self.tree_condensed

    def _tree_repr(self, *,
                   level: int=0,
                   max_level: int=None,
                   info_fstr="<{:cls_name,info}>",
                   info_ratio: float=0.6,
                   condense_thresh: Union[int, Callable[[int, int], int]]=None,
                   total_item_count: int=0
                   ) -> Union[str, List[str]]:
        """Recursively creates a multi-line string tree representation of this
        group. This is used by, e.g., the _format_tree method.
        
        Args:
            level (int, optional): The depth within the tree
            max_level (int, optional): The maximum depth within the tree;
                recursion is not continued beyond this level.
            info_fstr (str, optional): The format string for the info string
            info_ratio (float, optional): The width ratio of the whole line
                width that the info string takes
            condense_thresh (Union[int, Callable[[int, int], int]], optional):
                If given, this specifies the threshold beyond which the tree
                view for the current element becomes condensed by hiding the
                output for some elements.
                The minimum value for this is 3, indicating that there should
                be at most 3 lines be generated from this level (excluding the
                lines coming from recursion), i.e.: two elements and one line
                for indicating how many values are hidden.
                If a smaller value is given, this is silently brought up to 3.
                Half of the elements are taken from the beginning of the
                item iteration, the other half from the end.
                If given as integer, that number is used.
                If a callable is given, the callable will be invoked with the
                current level, number of elements to be added at this level,
                and the current total item count along this recursion branch.
                The callable should then return the number of lines to be
                shown for the current element.
            total_item_count (int, optional): The total number of items
                already created in this recursive tree representation call.
                Passed on between recursive calls.
        
        Returns:
            Union[str, List[str]]: The (multi-line) tree representation of
                this group. If this method was invoked with ``level == 0``, a
                string will be returned; otherwise, a list of strings will be
                returned.
        """
        def get_offset_str(level: int) -> str:
            """Returns an offst string, depending on level"""
            return "   " * level

        def truncate(s: str, *, max_length: int, suffix: str="…") -> str:
            """Truncates the given string to the desired length"""
            return s if len(s) <= max_length else s[:max_length-1] + suffix

        # Offset
        offset = get_offset_str(level)

        # Mark symbols
        first_mark = r' └┬'
        base_mark =  r'  ├'
        last_mark =  r'  └'
        only_mark =  r' └─'
        
        # Evaluate the condensation threshold, i.e. the maximum number of lines
        # to allow originating from this object (excluding recursion)
        num_items = len(self)
        total_item_count += num_items
        num_skipped = 0
        
        if callable(condense_thresh):
            max_lines = condense_thresh(level=level, num_items=num_items,
                                        total_item_count=total_item_count)
        else:
            max_lines = condense_thresh
        
        if max_lines is not None:
            # Additional check for lower bound; makes visualization much easier
            max_lines = max(3, int(max_lines))

            # If there are too few items, the variable is set to None to
            # indicate regular behavior.
            if num_items - max_lines < 1:
                max_lines = None

        # Calculations that make the output line fit into one terminal line
        num_cols = TTY_COLS
        info_width = int(num_cols * info_ratio)
        name_width = (num_cols - info_width) - (len(offset) + 3 + 1 + 2)

        def get_mark(n: int, *, max_n: int) -> str:
            """Returns the mark symbol depending on the iteration number.

            NOTE This uses variables from the outer scope!
            """
            if n == 0:
                if max_n == 0:
                    return only_mark
                return first_mark
            elif n == max_n:
                return last_mark
            return base_mark

        # The format string that's used to compose the whole output line
        fstr = "{offset:}{mark:>3s} {name:<{name_width}s}  {info:}"
        
        # Create the list to gather the lines in; add a description on level 0
        lines = []
        if level == 0:
            lines.append("")
            lines.append("Tree of {:logstr,info}".format(self))

        # Go over the entries on this level and format the lines
        for n, (key, obj) in enumerate(self.items()):
            # Determine whether to show this line of the tree or not. The lines
            # in the middle of the iteration are not shown.
            # If it is not shown, the first line that is then to be shown also
            # adds a line that indicates how many items were skipped.
            if max_lines is not None:
                if max_lines//2 <= n < (num_items - (max_lines-1)//2):
                    num_skipped += 1
                    continue

                elif n == (num_items - (max_lines-1)//2):
                    # Add the indicator line
                    lines.append(fstr.format(offset=offset, mark=base_mark,
                                             name_width=name_width,
                                             name="...",
                                             info=("... ({:d} more) ..."
                                                   "".format(num_skipped))))

            # Get key and info, truncating if necessary, and the mark
            name = truncate(key, max_length=name_width)
            info = truncate(info_fstr.format(obj), max_length=info_width)
            mark = get_mark(n, max_n=num_items-1)

            # Format the line and add to list of lines
            lines.append(fstr.format(offset=offset, mark=mark,
                                     name_width=name_width, name=name,
                                     info=info))

            # If it was a group and it is not empty...
            if isinstance(obj, BaseDataGroup) and len(obj) > 0:
                # ... and maximum recursion depth is not reached:
                if max_level is None or level < max_level:
                    # Continue recursion
                    lines += obj._tree_repr(level=level+1,
                                            max_level=max_level,
                                            info_fstr=info_fstr,
                                            info_ratio=info_ratio,
                                            condense_thresh=condense_thresh,
                                            total_item_count=total_item_count)

                else:
                    # Only indicate that it _would_ continue here, but do not
                    # actually continue with the recursion.
                    lines.append(fstr.format(offset=get_offset_str(level+1),
                                             mark=only_mark, name_width=3,
                                             name="...", info=""))

        # Done, depending on whether this is within the recursion or not,
        # return as list of lines or as combined multi-line string
        if level > 0:
            return lines
        return "\n".join(lines) + "\n"



"""Private low-level helper classes and functions for the DAG framework

NOTE This is imported by dantro.tools to register classes with YAML.
"""

from typing import Any, Union


# -----------------------------------------------------------------------------

class DAGReference:
    """The DAGReference class is the base class of all DAG reference objects.

    While it does not implement __hash__ by itself, it is yaml-representable
    and thus hashable after a parent object created a YAML representation.
    """
    def __init__(self, ref: str):
        """Initialize a DAGReference object from a hash."""
        if not isinstance(ref, str):
            raise TypeError("DAGReference requires a string-like argument, "
                            "got {}!".format(type(ref)))

        self._data = ref

    def __eq__(self, other) -> bool:
        """Only objects with exactly the same type and data are regarded as
        equal; specifically, this makes instances of subclasses always unequal
        to instances of the DAGReference base class.
        """
        if type(other) == type(self):
            return self._data == other._data
        return False

    def __repr__(self) -> str:
        return "<{} {}>".format(type(self).__name__, repr(self._data))

    def __hash__(self) -> int:
        return hash(repr(self))

    @property
    def ref(self) -> str:
        """The associated reference of this object"""
        return self._data

    def _resolve_ref(self, *, dag: 'TransformationDAG') -> str:
        """Return the hash reference; for the base class, the data is already
        the hash reference, so no DAG is needed. Derived classes _might_ need
        the DAG to resolve their reference hash.
        """
        return self._data

    def convert_to_ref(self, *, dag: 'TransformationDAG') -> 'DAGReference':
        """Create a new object that is a hash ref to the same object this
        tag refers to."""
        return DAGReference(self._resolve_ref(dag=dag))

    def resolve_object(self, *, dag: 'TransformationDAG') -> Any:
        """Resolve the object by looking up the reference in the DAG's object
        database.
        """
        return dag.objects[self._resolve_ref(dag=dag)]

    # YAML representation . . . . . . . . . . . . . . . . . . . . . . . . . . .
    yaml_tag = u'!dag_ref'

    @classmethod
    def from_yaml(cls, constructor, node):
        """Construct a DAGReference from a scalar YAML node"""
        return cls(constructor.construct_scalar(node))

    @classmethod
    def to_yaml(cls, representer, node):
        """Create a YAML representation of a DAGReference, carrying only the
        _data attribute over...
        
        As YAML expects scalar data to be str-like, a type cast is done. The
        subclasses that rely on certain argument types should take care that
        they can parse arguments that are str-like.
        """
        return representer.represent_scalar(cls.yaml_tag, str(node._data))


# .............................................................................

class DAGTag(DAGReference):
    """A DAGTag object stores a name of a tag, which serves as a named
    reference to some object in the DAG.

    While it does not implement __hash__ by itself, it is yaml-representable
    and thus hashable after a parent object created a YAML representation.
    """
    yaml_tag = u'!dag_tag'

    def __init__(self, name: str):
        """Initialize a DAGTag object, storing the specified field name"""
        self._data = name

    @property
    def name(self) -> str:
        """The name of the tag within the DAG that this object references"""
        return self._data

    def _resolve_ref(self, *, dag: 'TransformationDAG') -> str:
        """Return the hash reference by looking up the tag in the DAG"""
        return dag.tags[self.name]

# .............................................................................

class DAGNode(DAGReference):
    """A DAGNode is a reference by the index within the DAG's node list.

    While it does not implement __hash__ by itself, it is yaml-representable
    and thus hashable after a parent object created a YAML representation.
    """
    yaml_tag = u'!dag_node'

    def __init__(self, idx: int):
        """Initialize a DAGNode object with a node index.
        
        Args:
            idx (int): The idx value to set this reference to. Can also be a
                negative value, in which case the node list is traversed from
                the back.
        
        Raises:
            TypeError: On invalid type (not int-convertible)
        """
        if not isinstance(idx, int):
            # Try an integer conversion, to be a bit more robust
            try:
                idx = int(idx)
            except:
                raise TypeError("DAGNode requires an int-convertible "
                                "argument, got {}!".format(type(idx)))

        self._data = idx

    @property
    def idx(self) -> int:
        """The idx to the referenced node within the DAG's node list"""
        return self._data

    def _resolve_ref(self, *, dag: 'TransformationDAG') -> str:
        """Return the hash reference by looking up the node index in the DAG"""
        return dag.nodes[self.idx]


# -----------------------------------------------------------------------------

class DAGObjects:
    """An objects database for the DAG framework.

    It uses a flat dict containing (hash, object ref) pairs. The interface is
    slightly restricted compared to a regular dict; especially, item deletion
    is not made available.

    Objects are added to the database via the ``add_object`` method. They need
    to have a ``hashstr`` property, which returns a hash string
    deterministically representing the object; note that this is not
    equivalent to the Python builtin hash() function which invokes the __hash__
    magic method.
    """

    def __init__(self):
        """Initialize an empty objects database"""
        self._d = dict()

    def __str__(self) -> str:
        """A human-readable string representation of the object database"""
        return ("<DAGObjects database with {:d} entr{}>"
                "".format(len(self), "ies" if len(self) != 1 else "y"))

    def add_object(self, obj, *, custom_hash: str=None) -> str:
        """Add an object to the object database, storing it under its hash.
        
        Note that the object cannot be just any object that is hashable but it
        needs to return a string-based hash via the ``hashstr`` property. This
        is a dantro DAG framework-internal interface.
        
        Also note that the object will NOT be added if an object with the same
        hash is already present. The object itself is of no importance, only
        the returned hash is.
        
        Args:
            obj: Some object that has the ``hashstr`` property, i.e. is
                hashable as required by the DAG interface
            custom_hash (str, optional): A custom hash to use instead of the
                hash extracted from ``obj``. Can only be given when ``obj``
                does *not* have a ``hashstr`` property.
        
        Returns:
            str: The hash string of the given object. If a custom hash string
                was given, it is also the return value
        
        Raises:
            TypeError: When attempting to pass ``custom_hash`` while ``obj``
                *has* a ``hashstr`` property
            ValueError: If the given ``custom_hash`` already exists.
        """
        if custom_hash is not None:
            if hasattr(obj, 'hashstr'):
                raise TypeError("Cannot use a custom hash for objects that "
                                "provide their own `hashstr` property! Got "
                                "object of type {} and custom hash '{}'."
                                "".format(type(obj), custom_hash))

            elif custom_hash in self:
                raise ValueError("The provided custom hash '{}' for object of "
                                 "type {} already exists! Refusing to add it. "
                                 "Was the object already added? If not, "
                                 "choose a different custom hash."
                                 "".format(custom_hash, type(obj)))
            key = custom_hash
        
        else:
            # Use the DAG framework's internal hash method
            key = obj.hashstr

        # Only add the new object, if the hash does not exist yet.
        if key not in self:
            self._d[key] = obj
        return key

    def __getitem__(self, key: str) -> object:
        """Return the object associated with the given hash"""
        return self._d[key]

    def __len__(self) -> int:
        """Returns the number of objects in the objects database"""
        return len(self._d)

    def __contains__(self, key: str) -> bool:
        """Whether the given hash refers to an object in this database"""
        return key in self._d

    def keys(self):
        return self._d.keys()

    def values(self):
        return self._d.values()

    def items(self):
        return self._d.items()

# -----------------------------------------------------------------------------

def parse_dag_minimal_syntax(params: Union[str, dict]) -> dict:
    """Parses the minimal syntax parameters, effectively translating a string-
    like argument to a dict with the string specified as the ``operation`` key.
    """
    if isinstance(params, dict):
        return params

    elif isinstance(params, str):
        return dict(operation=params, with_previous_result=True)

    # else:
    raise TypeError("Expected either dict or string for minimal syntax, got "
                    "{} with value: {}".format(type(params), params))


def parse_dag_syntax(*, operation: str=None, args: list=None,
                     kwargs: dict=None, tag: str=None,
                     with_previous_result: bool=False,
                     salt: int=None, file_cache: dict=None, **ops) -> dict:
    """Given the parameters of a transform operation, possibly in a shorthand
    notation, returns a dict with normalized content by expanding the
    shorthand notation.
    
    Keys that will be available in the resulting dict:
        ``operation``, ``args``, ``kwargs``, ``tag``.
    
    Args:
        operation (str, optional): Which operation to carry out; can only be
            specified if there is no ``ops`` argument.
        args (list, optional): Positional arguments for the operation; can
            only be specified if there is no ``ops`` argument.
        kwargs (dict, optional): Keyword arguments for the operation; can only
            be specified if there is no ``ops`` argument.
        tag (str, optional): The tag to attach to this transformation
        with_previous_result (bool, optional): Whether the result of the
            previous transformation is to be used as first positional argument
            of this transformation.
        salt (int, optional): A salt to the Transformation object, thereby
            changing its hash.
        file_cache (dict, optional): File cache parameters
        **ops: The operation that is to be carried out. May contain one and
            only one operation.
    
    Returns:
        dict: The normalized dict of transform parameters.
    
    Raises:
        ValueError: For len(ops) != 1
    """
    # Distinguish between explicit and shorthand mode
    if operation and not ops:
        # Explicit parametrization
        args = args if args else []
        kwargs = kwargs if kwargs else {}

    elif ops and not operation:
        # Shorthand parametrization
        # Make sure there are no stray argument
        if args is not None or kwargs is not None:
            raise ValueError("When using shorthand notation, the args and "
                             "kwargs need to be specified under the key that "
                             "specifies the operation!")

        elif len(ops) > 1:
            raise ValueError("For shorthand notation, there can only be a "
                             "single operation specified, but got: {}."
                             "".format(ops))

        # Extract operation name and parameters
        operation, op_params = list(ops.items())[0]

        # Depending on type, regard parameters as args or kwargs. If
        # the argument is not a container, assume it's a single
        # positional argument.
        if isinstance(op_params, dict):
            args, kwargs = [], op_params
        
        elif isinstance(op_params, (list, tuple)):
            args, kwargs = list(op_params), {}
        
        elif op_params is not None:
            args, kwargs = [op_params], {}
        
        else:
            args, kwargs = [], {}

    elif not operation and not ops:
        raise ValueError("Missing operation specification. Either use the "
                         "`operation` key to specify one or use shorthand "
                         "notation by using the name of the operation as a "
                         "key and adding the arguments to it as values.")

    else:
        raise ValueError("Got two specifications of operations, one via the "
                         "`operation` argument ({}), another via the "
                         "shorthand notation ({}). Remove one of them."
                         "".format(operation, ops))

    # Have variables operation, args, and kwargs set now.

    # If the result is to be carried on, the first _positional_
    # argument is set to be a reference to the previous node
    if with_previous_result:
        args.insert(0, DAGNode(-1))

    # Done. Construct the dict.
    # Mandatory parameters
    d = dict(operation=operation, args=args, kwargs=kwargs, tag=tag)
    
    # Add optional parameters only if they were specified
    if salt is not None:
        d['salt'] = salt

    if file_cache is not None:
        d['file_cache'] = file_cache

    return d



"""In this module, the NetworkGroup is implemented"""

import logging
from typing import List, Union

import numpy as np
import xarray as xr
import networkx as nx
import networkx.exception

from . import TimeSeriesGroup
from ..base import BaseDataGroup
from ..containers import XrDataContainer

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class NetworkGroup(BaseDataGroup):
    """The NetworkGroup class manages groups of network data containers."""

    # Define allowed member container types
    _ALLOWED_CONT_TYPES = (XrDataContainer, TimeSeriesGroup)

    # Define, as class variables, in which containers or attributes to find the
    # info on the nodes and edges.
    _NWG_node_container = "nodes"
    _NWG_edge_container = "edges"
    _NWG_attr_directed = "directed"
    _NWG_attr_parallel = "parallel"
    _NWG_attr_node_prop = "node_prop"
    _NWG_attr_edge_prop = "edge_prop"

    # .........................................................................

    def _get_data_at(self, *, data: Union[XrDataContainer, TimeSeriesGroup],
                     at_time: int=None, at_time_idx: int=None
                     ) -> Union[xr.DataArray, XrDataContainer]:
        """Extract the data (at a certain time).
        
        Args:
            data (Union[XrDataContainer, TimeSeriesGroup]): data to select from
            at_time (int, optional): access data via time coordinate
            at_time_idx (int, optional): access data via index
        
        Returns:
            Union[xr.DataArray, XrDataContainer]: selected property data
        
        Raises:
            IndexError: selected time index not available
            KeyError: selected time coordinate not available
            TypeError: wrong container for single graph data
            ValueError: Both ``at_time`` and ``at_time_idx`` are given
        
        """
        if at_time_idx is not None and at_time is not None:
            raise ValueError("Arguments 'at_time' and 'at_time_idx' can not "
                             "be given at the same time!")

        # If `time` is not available, assume 1d data and return as it is.
        if 'time' not in data.dims:
            return data

        if at_time is not None:

            try:
                data = data.sel(time=at_time)

            except KeyError as err:
                raise KeyError("Time key '{}' not available!"
                               "".format(at_time)) from err

        elif at_time_idx is not None:

            try:
                data = data.isel(time=at_time_idx)

            except IndexError as err:
                raise IndexError("Time index '{}' not available!"
                                 "".format(at_time_idx)) from err

        else:
            if not isinstance(data, XrDataContainer):
                raise TypeError("'data' has to be a XrDataContainer"
                                ", got '{}'".format(type(data)))

        return data


    def create_graph(self, *, directed: bool=None, parallel_edges: bool=None,
                     at_time: int=None, at_time_idx: int=None,
                     node_props: list=None, edge_props: list=None,
                     **graph_kwargs) -> nx.Graph:
        """Create a networkx graph object.
        
        Args:
            directed (bool, optional): The graph is directed. If not given, the
                value given by the group attribute with name
                ``_NWG_attr_directed`` is used instead.
            parallel_edges (bool, optional): If true, the graph will allow
                parallel edges. If not given, the value is tried to be read
                from the group attribute with name _NWG_attr_parallel.
            at_time (int, optional): access data via time coordinate
            at_time_idx (int, optional): access data via index
            node_props (list, optional): list of names specifying the
                containers that contain the node property data
            edge_props (list, optional): list of names specifying the
                containers that contain the edge property data
            **graph_kwargs: Further initialization kwargs for the graph.
        
        Returns:
            nx.Graph
        
        Raises:
            KeyError: For non-existant node or edge container under the names
                specified by the respective class variables.
        """

        # store nodes and edges here
        node_cont = None
        edge_cont = None

        # Check types and extract data as it is
        log.debug("Checking whether node and edge container are available...")
        
        try:
            node_cont = self[self._NWG_node_container]

        except KeyError as err:
            raise KeyError("No container with name '{}' available in {}! "
                           "Check if the class variable _NWG_node_container "
                           "is set to the correct value."
                           "".format(self._NWG_node_container, self.logstr)
                           ) from err

        try:
            edge_cont = self[self._NWG_edge_container]

        except KeyError as err:
            raise KeyError("No container with name '{}' available in {}! "
                           "Check if the class variable _NWG_edge_container "
                           "is set to the correct value."
                           "".format(self._NWG_edge_container, self.logstr)
                           ) from err

        # Get node and edge data. If only data for a single time step was
        # written, 'at_time' and 'at_time_idx' are ignored.
        if ('time' in node_cont.dims and
            len(node_cont.coords['time']) == 1):
            node_cont = self._get_data_at(data=node_cont, at_time_idx=0)
        else:
            node_cont = self._get_data_at(data=node_cont, at_time=at_time,
                                            at_time_idx=at_time_idx)

        if ('time' in edge_cont.dims and
            len(edge_cont.coords['time']) == 1):
            edge_cont = self._get_data_at(data=edge_cont, at_time_idx=0)
        else:
            edge_cont = self._get_data_at(data=edge_cont, at_time=at_time,
                                            at_time_idx=at_time_idx)

        # Get info on directed and parallel edges from attributes, if not
        # explicitly given
        if directed is None:
            directed = self.attrs[self._NWG_attr_directed]
        
        if parallel_edges is None:
            parallel_edges = self.attrs[self._NWG_attr_parallel]

        # sort nodes
        node_cont = node_cont[np.argsort(node_cont)]

        # NOTE: Only the nodes and the node data has to be sorted as the edges
        #       are defined via the node id

        # Create a networkx graph corresponding to the graph properties.
        log.debug("Creating a networkx graph object...")
        
        if not directed and not parallel_edges:
            g = nx.Graph(**graph_kwargs)
            
        elif directed and not parallel_edges:
            g = nx.DiGraph(**graph_kwargs)
            
        elif not directed and parallel_edges:
            g = nx.MultiGraph(**graph_kwargs)
            
        else:
            g = nx.MultiDiGraph(**graph_kwargs)

        # Add nodes to the graph
        log.debug("Adding nodes to the graph...")
        g.add_nodes_from(node_cont.values)
        
        # Add edges to the graph
        log.debug("Adding edges to the graph...")
        try:
            g.add_edges_from(edge_cont.values)
        
        except networkx.exception.NetworkXError as err:
            # Probably the data had an unexpected shape. 
            # Just try transposing :)
            try:
                g.add_edges_from(edge_cont.values.T)
            
            except:
                # It doesn't work. Give up on it... :(
                raise err


        # Add properties to nodes and edges
        # The properties will be sorted along the node id's
        if node_props:
            for prop_name in node_props:
                self.set_node_property(g=g, name=prop_name,
                            at_time=at_time, at_time_idx=at_time_idx)
        if edge_props:
            for prop_name in edge_props:
                self.set_edge_property(g=g, name=prop_name,
                            at_time=at_time, at_time_idx=at_time_idx)

        # Return the graph
        log.info("Successfully created graph (%d node%s, %d edge%s) from %s.",
                 g.number_of_nodes(), "s" if g.number_of_nodes() != 1 else "",
                 g.number_of_edges(), "s" if g.number_of_edges() != 1 else "",
                 self.logstr)

        return g


    def set_node_property(self, *, g: nx.Graph, name: str, at_time: int=None,
                          at_time_idx: int=None):
        """Sets a property to every node in Graph g.
        
        Args:
            g (nx.Graph): The networkx Graph to work on
            name (str): Name of the container which contains the property
                values
            at_time (int, optional): access data via time coordinate
            at_time_idx (int, optional): access data via index
        
        Raises:
            AttributeError: specified container not marked as property via
                the according class attribute
            KeyError: The name is not available in ``nwgrp`` or ``at_time`` is
                not available in ``nwgrp[name]``.
            ValueError: lenght of datasets does not match
        """

        try:
            prop_data = self[name]

        except KeyError as err:
            raise KeyError("No container with name '{}' available in"
                           " {}!".format(name, self.logstr)) from err

        if not prop_data.attrs.get(self._NWG_attr_node_prop, False):
            raise AttributeError("The data in '{}' is not marked as node"
                " property! Check that it has the attribute '{}' set to"
                " True".format(name, self._NWG_attr_node_prop))

        prop_data = self._get_data_at(data=prop_data, at_time=at_time,
                                        at_time_idx=at_time_idx)

        # sort data along the node id's
        node_cont = self[self._NWG_node_container]

        if ('time' in node_cont.dims
            and len(node_cont.coords['time']) == 1):
            node_cont = self._get_data_at(data=node_cont, at_time_idx=0)
        else:
            node_cont = self._get_data_at(data=node_cont, at_time=at_time,
                                            at_time_idx=at_time_idx)
            
        prop_data = prop_data[np.argsort(node_cont)]

        if len(g.nodes) != len(prop_data.values):
            raise ValueError("Mismatch! '{}' property values for '{}' nodes!"
                             "".format(len(prop_data.values), len(g.nodes)))

        props = dict()

        for node, val in zip(g.nodes, prop_data.values):
            props[node] = val

        # add property to graph
        nx.set_node_attributes(g, props, name=name)
        

    def set_edge_property(self, *, g: nx.Graph, name: str, at_time: int=None,
                          at_time_idx: int=None):
        """Sets a property to every edge in Graph g.
        
        Args:
            g (nx.Graph): The networkx Graph to work on
            name (str): Name of the container which contains the property
                values
            at_time (int, optional): access data via time coordinate
            at_time_idx (int, optional): access data via index
        
        Raises:
            AttributeError: specified container not marked as property via
                the according class attribute
            KeyError: The name is not available in ``nwgrp`` or ``at_time`` is
                not available in ``nwgrp[name]``.
            ValueError: dataset lengths mismatch
        """
        try:
            prop_data = self[name]

        except KeyError as err:
            raise KeyError("No container with name '{}' available in {}!"
                           "".format(name, self.logstr)) from err

        if not prop_data.attrs.get(self._NWG_attr_edge_prop, False):
            raise AttributeError("The data in '{}' is not marked as edge "
                                 "property! Check that it has the attribute "
                                 "'{}' set to True"
                                 "".format(name, self._NWG_attr_edge_prop))

        prop_data = self._get_data_at(data=prop_data, at_time=at_time,
                                      at_time_idx=at_time_idx)

        if len(g.edges) != len(prop_data.values):
            raise ValueError("Mismatch! '{}' property values for '{}' edges"
                             "".format(len(prop_data.values), len(g.edges)))

        props = dict()
        for edge, val in zip(g.edges, prop_data.values):
            props[edge] = val

        # add property to graph
        nx.set_edge_attributes(g, props, name=name)



"""Defines a loader mixin to load numpy dumps"""

import numpy as np

from ..containers import NumpyDataContainer
from ._tools import add_loader

# -----------------------------------------------------------------------------

class NumpyLoaderMixin:
    """Supplies functionality to load numpy binary dumps into numpy objects"""

    @add_loader(TargetCls=NumpyDataContainer)
    def _load_numpy_binary(filepath: str, *, TargetCls: type,
                           **load_kwargs) -> NumpyDataContainer:
        """Loads the output of ``numpy.save`` back into a
        :py:class:`~dantro.containers.numeric.NumpyDataContainer`.
        
        Args:
            filepath (str): Where the ``*.npy`` file is located
            TargetCls (type): The class constructor
            **load_kwargs: Passed on to ``numpy.load``, see there for kwargs
        
        Returns:
            NumpyDataContainer: The reconstructed NumpyDataContainer
        """
        return TargetCls(data=np.load(filepath, **load_kwargs),
                         attrs=dict(filepath=filepath))

    # Also make it available under plain 'numpy'
    _load_numpy = _load_numpy_binary



"""Takes care of all YAML-related imports and configuration"""

import os
import io
import logging
from typing import Any

import ruamel.yaml

# Local constants
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

# Import yaml from paramspace and configure it
from paramspace import yaml
yaml.default_flow_style = False

# Register further classes
from ._dag_utils import DAGReference, DAGTag, DAGNode

yaml.register_class(DAGReference)
yaml.register_class(DAGTag)
yaml.register_class(DAGNode)

# Special constructors ........................................................
# For the case of a reference to the previous node
yaml.constructor.add_constructor(u'!dag_prev', lambda l, n: DAGNode(-1))

# -----------------------------------------------------------------------------

def load_yml(path: str, *, mode: str='r') -> dict:
    """Loads a yaml file from a path
    
    Args:
        path (str): The path to the yml file
        mode (str, optional): Read mode
    
    Returns:
        dict: The parsed dictionary data
    """
    log.debug("Loading YAML file... mode: %s, path:\n  %s", mode, path)

    with open(path, mode) as yaml_file:
        return yaml.load(yaml_file)

def write_yml(d: dict, *, path: str, mode: str='w'):
    """Write a dict as a yaml file to a path
    
    Args:
        d (dict): The dict to convert to dump
        path (str): The path to write the yml file to
        mode (str, optional): Write mode of the file
    """
    log.debug("Dumping %s to YAML file... mode: %s, target:\n  %s",
              type(d).__name__, mode, path)

    # Make sure the directory is present
    os.makedirs(os.path.dirname(path), exist_ok=True)

    with open(path, mode) as yaml_file:
        # Add the yaml '---' prefix
        yaml_file.write("---\n")

        # Now dump the rest
        yaml.dump(d, stream=yaml_file)

        # Ensure new line
        yaml_file.write("\n")


# -----------------------------------------------------------------------------
# More specific YAML-based functions

def yaml_dumps(obj: Any, *, register_classes: tuple=(), **dump_params) -> str:
    """Serializes the given object using a newly created YAML dumper.
    
    The aim of this function is to provide YAML dumping that is not dependent
    on any package configuration; all parameters can be passed here.
    
    In other words, his function does _not_ use the dantro._yaml.yaml object
    for dumping but each time creates a new dumper with fixed settings. This
    reduces the chance of interference from elsewhere. Compared to the time
    needed for serialization in itself, the extra time needed to create the
    new ruamel.yaml.YAML object and register the classes is negligible.
    
    Args:
        obj (Any): The object to dump
        register_classes (tuple, optional): Additional classes to register
        **dump_params: Dumping parameters
    
    Returns:
        str: The output of serialization
    
    Raises:
        ValueError: On failure to serialize the given object
    """
    s = io.StringIO()
    y = ruamel.yaml.YAML()

    # Register classes; then apply dumping parameters via object properties
    for Cls in register_classes:
        y.register_class(Cls)

    for k, v in dump_params.items():
        setattr(y, k, v)

    # Serialize
    try:
        y.dump(obj, stream=s)
    
    except Exception as err:
        raise ValueError("Could not serialize the given {} object!"
                         "".format(type(obj))) from err

    return s.getvalue()



"""This module implements mixin classes which provide numeric interfaces for
containers
"""

import math
import operator
import logging

import numpy as np

from ..abc import AbstractDataContainer

# Local variables
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class UnaryOperationsMixin:
    """This Mixin class implements the methods needed for unary operations.
    
    It leaves out those that expect that return values are of a certain type,
    e.g. __complex__, __int__, ...
    """

    def __neg__(self):
        """Make negative
        
        Returns:
            A new object with negative elements
        """
        return apply_func_to_copy(self, operator.neg)

    def __pos__(self):
        """Make positive
        
        Returns:
            A new object with negative elements
        """
        return apply_func_to_copy(self, operator.pos)

    def __abs__(self):
        """Absolute value

        Returns:
            A new object with the absolute value of the elements
        """
        return apply_func_to_copy(self, operator.abs)

    def __invert__(self):
        """Inverse value

        Returns:
            A new object with the inverted values of the elements
        """
        return apply_func_to_copy(self, operator.invert)

    def __round__(self):
        """Rounds number to nearest integer

        Returns:
            A new object as rounded number to nearest integer
        """
        return apply_func_to_copy(self, round)

    def __ceil__(self):
        """Smallest integer

        Returns:
            A new object containing the smallest integer
        """
        return apply_func_to_copy(self, math.ceil)

    def __floor__(self):
        """Largest integer

        Returns:
            A new object containing the largest element
        """
        return apply_func_to_copy(self, math.floor)

    def __trunc__(self):
        """Truncated to the nearest integer toward 0

        Returns:
            A new object containing the truncated element
        """
        return apply_func_to_copy(self, math.trunc)


class NumbersMixin(UnaryOperationsMixin):
    """This mixin implements the methods needed for calculating with numbers.
    """

    def __add__(self, other):
        """Add two objects
        
        Returns:
            A new object containing the summed data
        """
        return apply_func_to_copy(self, operator.add, other)

    def __sub__(self, other):
        """Subtract two objects
        
        Returns:
            A new object containing the subtracted data
        """
        return apply_func_to_copy(self, operator.sub, other)

    def __mul__(self, other):
        """Multiply two objects
        
        Returns:
            A object containing the multiplied data
        """
        return apply_func_to_copy(self, operator.mul, other)

    def __truediv__(self, other):
        """Divide two objects
        
        Returns:
            A new object containing the divided data
        """
        return apply_func_to_copy(self, operator.truediv, other)

    def __floordiv__(self, other):
        """Floor divide two objects
        
        Returns:
            A new object containing the floor divided data
        """
        return apply_func_to_copy(self, operator.floordiv, other)

    def __mod__(self, other):
        """Calculate the modulo of two objects
        
        Returns:
            A new object containing the summed data
        """
        return apply_func_to_copy(self, operator.mod, other)

    def __divmod__(self, other):
        """Calculate the floor division and modulo of two objects
        
        Returns:
            A new object containing the floor divided data and its modulo
        """
        return apply_func_to_copy(self, divmod, other)

    def __pow__(self, other):
        """Calculate the self data to the power of other data
        
        Returns:
            A new object containing the result
        """
        return apply_func_to_copy(self, operator.pow, other)

    # inplace operations
    def __iadd__(self, other):
        """Add two objects
        
        Returns:
            Self with modified data
        """
        return apply_func_inplace(self, operator.iadd, other)

    def __isub__(self, other):
        """Subtract two objects
        
        Returns:
            Self with modified data
        """
        return apply_func_inplace(self, operator.isub, other)

    def __imul__(self, other):
        """Multiply two objects
        
        Returns:
            Self with modified data
        """
        return apply_func_inplace(self, operator.imul, other)

    def __itruediv__(self, other):
        """Divide two objects
        
        Returns:
            Self with modified data
        """
        return apply_func_inplace(self, operator.itruediv, other)

    def __ifloordiv__(self, other):
        """Floor divide two objects
        
        Returns:
            Self with modified data
        """
        return apply_func_inplace(self, operator.ifloordiv, other)

    def __imod__(self, other):
        """Calculate the modulo of two objects
        
        Returns:
            Self with modified data
        """
        return apply_func_inplace(self, operator.imod, other)

    def __ipow__(self, other):
        """Calculate the self data to the power of other data
        
        Returns:
            Self with modified data
        """
        return apply_func_inplace(self, operator.ipow, other)


class ComparisonMixin:
    """This Mixin implements functions to compare objects"""

    def __eq__(self, other):
        """Equality"""
        return self.data == get_data(other)

    def __ne__(self, other):
        """Inequality"""
        return self.data != get_data(other)
    
    def __lt__(self, other):
        """Less than"""
        return self.data < get_data(other)

    def __le__(self, other):
        """Less than or equal"""
        return self.data <= get_data(other)

    def __gt__(self, other):
        """Greater than"""
        return self.data > get_data(other)

    def __ge__(self, other):
        """Greater than or equal"""
        return self.data >= get_data(other)

    def __bool__(self):
        """Truth value"""
        return bool(self.data)


# -----------------------------------------------------------------------------
# Helpers ---------------------------------------------------------------------
# -----------------------------------------------------------------------------

def get_data(obj):
    """Get the data of `obj` depending on whether it is part of dantro or not.
    
    Args:
        obj: The object to check
    
    Returns:
        Either the `.data` attribute of a dantro-based object or otherwise the
            object itself.
    """
    if isinstance(obj, AbstractDataContainer):
        return obj.data
    # Not dantro-based, just return the object itself.
    return obj


def apply_func_to_copy(obj, func, other=None):
    """Apply a given function to a copy for all datatypes
    
    Returns:
        An object with the data on which the function was applied
    """
    # Work on a copy
    new = obj.copy()

    # Change the data of the new object
    if other is None:
        new._data = func(new.data)
    else:
        if isinstance(other, AbstractDataContainer):
            new._data = func(new.data, other.data)
        else:
            new._data = func(new.data, other)

    return new


def apply_func_inplace(obj, func, other=None):
    """Apply a given function inplace for all datatypes
    
    Returns:
        An object with the data on which the function was applied
    """
    # Change the data of the new object
    if other is None:
        func(obj._data)
    else:
        if isinstance(other, AbstractDataContainer):
            func(obj._data, other.data)
        else:
            func(obj._data, other)
            
    return obj




import pickle as pkl

from ..containers import ObjectContainer
from ._tools import add_loader

# -----------------------------------------------------------------------------

class PickleLoaderMixin:
    """Supplies functionality to load pickled python objects"""

    # Define the load function such that it can be set in a flexible way
    _PICKLE_LOAD_FUNC = pkl.load

    @add_loader(TargetCls=ObjectContainer, omit_self=False)
    def _load_pickle(self, filepath: str, *, TargetCls: type, **pkl_kwargs) -> ObjectContainer:
        """Load a pickled object.

        This uses the load function defined under the _PICKLE_LOAD_FUNC class
        variable, which defaults to the pickle.load function.
        
        Args:
            filepath (str): Where the pickle-dumped file is located
            TargetCls (type): The class constructor
            **pkl_kwargs: Passed on to the load function
        
        Returns:
            ObjectContainer: The unpickled file, stored in a dantro container
        """
        # Open file in binary mode and unpickle with the given load function
        with open(filepath, mode='rb') as f:
            obj = self._PICKLE_LOAD_FUNC(f, **pkl_kwargs)

        # Populate the target container with the object
        return TargetCls(data=obj, attrs=dict(filepath=filepath))

    # Also make available under `pkl`
    _load_pkl = _load_pickle



"""This module implements mixins that provide proxy support"""

import logging
import warnings
from typing import Union

import numpy as np

from ..abc import AbstractDataProxy

# Local variables
log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

class ProxySupportMixin:
    """This Mixin class overwrites the ``data`` property to allow and resolve
    proxy objects.

    It should be used to add support for certain proxy types to a container.

    A proxy object is a place holder for data that is not yet loaded. It will
    only be loaded if the ``data`` property is directly or indirectly accessed.
    """

    # If the CheckDataMixin is used, it should also allow proxy data
    DATA_ALLOW_PROXY = True
    # NOTE Depending on the order of how the mixins are given, this might
    # not have an effect. Make sure the proxy mixin is loaded after the
    # CheckDataMixin or the container class that uses the CheckDataMixin

    # Which type to resolve the proxy to
    PROXY_RESOLVE_ASTYPE = None

    # Whether to retain the proxy object after resolving
    PROXY_RETAIN = False

    # Behaviour upon failure of reinstating a proxy
    PROXY_REINSTATE_FAIL_ACTION = 'raise'  # or:  warn, log_warning, log_debug

    # Make sure the attribute where a retained proxy is stored is available
    _retained_proxy = None

    @property
    def data(self):
        """The container data. If the data is a proxy, this call will lead
        to the resolution of the proxy.
        
        Returns:
            The data stored in this container
        """
        # Have to check whether the data might be a proxy. If so, resolve it.
        if self.data_is_proxy:
            log.debug("Resolving %s for %s ...",
                      self._data.classname, self.logstr)

            # Optionally, retain the proxy object. If not doing this, will go
            # out of scope
            if self.PROXY_RETAIN:
                log.debug("...retaining proxy object...")
                self._retained_proxy = self._data

            # Now, overwrite the _data attribute where the proxy used to be
            self._data = self.proxy.resolve(astype=self.PROXY_RESOLVE_ASTYPE)

            # Postprocess the resolved proxy with optional method
            if hasattr(self, '_postprocess_proxy_resolution'):
                log.debug("Calling proxy resolution postprocessing ...")
                self._postprocess_proxy_resolution()

        # Now, the data should be loaded and can be returned
        return self._data

    @property
    def data_is_proxy(self) -> bool:
        """Returns true, if this is proxy data
        
        Returns:
            bool: Whether the *currently* stored data is a proxy object
        """
        return isinstance(self._data, AbstractDataProxy)

    @property
    def proxy(self) -> Union[AbstractDataProxy, None]:
        """If the data is proxy, returns the proxy data object without using 
        the .data attribute (which would trigger resolving the proxy); else 
        returns None.
        
        Returns:
            Union[AbstractDataProxy, None]: If the data is proxy, return the
                proxy object; else None.
        """
        if self.data_is_proxy:
            return self._data
        return self._retained_proxy

    def reinstate_proxy(self):
        """Re-instate a previously retained proxy object, discarding _data."""
        if self.data_is_proxy:
            return

        if self._retained_proxy is None:
            msg = ("Could not reinstate a proxy for {} because there was no "
                   "proxy retained. Was there one in the first place? Make "
                   "sure the `PROXY_RETAIN` class variable is set to True. "
                   "To control the behaviour of this message, change the "
                   "PROXY_REINSTATE_FAIL_ACTION class variable."
                   "".format(self.logstr))
            action = self.PROXY_REINSTATE_FAIL_ACTION
            
            if action == 'raise':
                raise ValueError(msg)

            elif action == 'warn':
                warnings.warn(msg, RuntimeWarning)

            elif action in ('log_warning', 'log_warn'):
                log.warning(msg)

            elif action == 'log_debug':
                log.debug(msg)

            else:
                raise ValueError("Invalid PROXY_REINSTATE_FAIL_ACTION value "
                                 "'{}'! Possible values: raise, warn, "
                                 "log_warning, log_debug".format(action))

        else:
            # All good. Reinstate the proxy
            self._data = self._retained_proxy
            log.debug("Reinstated %s for data of %s.",
                      self.proxy.classname, self.logstr)

    def _format_info(self) -> str:
        """Adds an indicator to whether data is proxy to the info string.
        Additionally, the proxy tags are appended.
        """
        if self.data_is_proxy:
            tags = (" ({})".format(", ".join(self.proxy.tags))
                    if self.proxy.tags else "")
            return "proxy{}, {}".format(tags, super()._format_info())
        return super()._format_info()


class Hdf5ProxySupportMixin(ProxySupportMixin):
    """Specializes the
    :py:class:`~dantro.mixins.proxy_support.ProxySupportMixin` to the
    capabilities of :py:class:`~dantro.proxy.hdf5.Hdf5DataProxy`, i.e. it
    allows access to the cached properties of the proxy object without
    resolving it.
    """

    @property
    def dtype(self) -> np.dtype:
        """Returns dtype, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.dtype
        return self.data.dtype
    
    @property
    def shape(self) -> tuple:
        """Returns shape, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.shape
        return self.data.shape

    @property
    def ndim(self) -> int:
        """Returns ndim, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.ndim
        return self.data.ndim

    @property
    def size(self) -> int:
        """Returns size, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.size
        return self.data.size

    @property
    def chunks(self) -> tuple:
        """Returns chunks, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.chunks
        return self.data.chunks
